# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Public TF-Keras utilities."""

# isort: off

# Serialization related
from tf_keras.src.saving.serialization_lib import deserialize_keras_object
from tf_keras.src.saving.serialization_lib import serialize_keras_object
from tf_keras.src.saving.object_registration import CustomObjectScope
from tf_keras.src.saving.object_registration import custom_object_scope
from tf_keras.src.saving.object_registration import get_custom_objects
from tf_keras.src.saving.object_registration import get_registered_name
from tf_keras.src.saving.object_registration import register_keras_serializable

# Dataset related
from tf_keras.src.utils.audio_dataset import audio_dataset_from_directory
from tf_keras.src.utils.text_dataset import text_dataset_from_directory
from tf_keras.src.utils.timeseries_dataset import timeseries_dataset_from_array
from tf_keras.src.utils.image_dataset import image_dataset_from_directory
from tf_keras.src.utils.dataset_utils import split_dataset

# Sequence related
from tf_keras.src.utils.data_utils import GeneratorEnqueuer
from tf_keras.src.utils.data_utils import OrderedEnqueuer
from tf_keras.src.utils.data_utils import Sequence
from tf_keras.src.utils.data_utils import SequenceEnqueuer

# Image related
from tf_keras.src.utils.image_utils import array_to_img
from tf_keras.src.utils.image_utils import img_to_array
from tf_keras.src.utils.image_utils import load_img
from tf_keras.src.utils.image_utils import save_img

# Python utils
from tf_keras.src.utils.tf_utils import set_random_seed
from tf_keras.src.utils.generic_utils import Progbar
from tf_keras.src.utils.data_utils import get_file

# Preprocessing utils
from tf_keras.src.utils.feature_space import FeatureSpace

# Internal
from tf_keras.src.utils.layer_utils import get_source_inputs
from tf_keras.src.utils.layer_utils import warmstart_embedding_matrix

# Deprecated
from tf_keras.src.utils.np_utils import normalize
from tf_keras.src.utils.np_utils import to_categorical
from tf_keras.src.utils.np_utils import to_ordinal
from tf_keras.src.utils.data_utils import pad_sequences

# Evaluation related
from tf_keras.src.utils.sidecar_evaluator import SidecarEvaluator
from tf_keras.src.utils.sidecar_evaluator import SidecarEvaluatorModelExport

# Timed Thread
from tf_keras.src.utils.timed_threads import TimedThread

# Visualization related
from tf_keras.src.utils.vis_utils import model_to_dot
from tf_keras.src.utils.vis_utils import plot_model




"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.backend import experimental
from tf_keras.src.backend import abs
from tf_keras.src.backend import all
from tf_keras.src.backend import any
from tf_keras.src.backend import arange
from tf_keras.src.backend import argmax
from tf_keras.src.backend import argmin
from tf_keras.src.backend import backend
from tf_keras.src.backend import batch_dot
from tf_keras.src.backend import batch_flatten
from tf_keras.src.backend import batch_get_value
from tf_keras.src.backend import batch_normalization
from tf_keras.src.backend import batch_set_value
from tf_keras.src.backend import bias_add
from tf_keras.src.backend import binary_crossentropy
from tf_keras.src.backend import binary_focal_crossentropy
from tf_keras.src.backend import cast
from tf_keras.src.backend import cast_to_floatx
from tf_keras.src.backend import categorical_crossentropy
from tf_keras.src.backend import categorical_focal_crossentropy
from tf_keras.src.backend import clear_session
from tf_keras.src.backend import clip
from tf_keras.src.backend import concatenate
from tf_keras.src.backend import constant
from tf_keras.src.backend import conv1d
from tf_keras.src.backend import conv2d
from tf_keras.src.backend import conv2d_transpose
from tf_keras.src.backend import conv3d
from tf_keras.src.backend import cos
from tf_keras.src.backend import count_params
from tf_keras.src.backend import ctc_batch_cost
from tf_keras.src.backend import ctc_decode
from tf_keras.src.backend import ctc_label_dense_to_sparse
from tf_keras.src.backend import cumprod
from tf_keras.src.backend import cumsum
from tf_keras.src.backend import depthwise_conv2d
from tf_keras.src.backend import dot
from tf_keras.src.backend import dropout
from tf_keras.src.backend import dtype
from tf_keras.src.backend import elu
from tf_keras.src.backend import equal
from tf_keras.src.backend import eval
from tf_keras.src.backend import exp
from tf_keras.src.backend import expand_dims
from tf_keras.src.backend import eye
from tf_keras.src.backend import flatten
from tf_keras.src.backend import foldl
from tf_keras.src.backend import foldr
from tf_keras.src.backend import function
from tf_keras.src.backend import gather
from tf_keras.src.backend import get_uid
from tf_keras.src.backend import get_value
from tf_keras.src.backend import gradients
from tf_keras.src.backend import greater
from tf_keras.src.backend import greater_equal
from tf_keras.src.backend import hard_sigmoid
from tf_keras.src.backend import in_test_phase
from tf_keras.src.backend import in_top_k
from tf_keras.src.backend import in_train_phase
from tf_keras.src.backend import int_shape
from tf_keras.src.backend import is_keras_tensor
from tf_keras.src.backend import is_sparse
from tf_keras.src.backend import l2_normalize
from tf_keras.src.backend import learning_phase
from tf_keras.src.backend import learning_phase_scope
from tf_keras.src.backend import less
from tf_keras.src.backend import less_equal
from tf_keras.src.backend import local_conv1d
from tf_keras.src.backend import local_conv2d
from tf_keras.src.backend import log
from tf_keras.src.backend import manual_variable_initialization
from tf_keras.src.backend import map_fn
from tf_keras.src.backend import max
from tf_keras.src.backend import maximum
from tf_keras.src.backend import mean
from tf_keras.src.backend import min
from tf_keras.src.backend import minimum
from tf_keras.src.backend import moving_average_update
from tf_keras.src.backend import name_scope
from tf_keras.src.backend import ndim
from tf_keras.src.backend import normalize_batch_in_training
from tf_keras.src.backend import not_equal
from tf_keras.src.backend import one_hot
from tf_keras.src.backend import ones
from tf_keras.src.backend import ones_like
from tf_keras.src.backend import permute_dimensions
from tf_keras.src.backend import placeholder
from tf_keras.src.backend import pool2d
from tf_keras.src.backend import pool3d
from tf_keras.src.backend import pow
from tf_keras.src.backend import print_tensor
from tf_keras.src.backend import prod
from tf_keras.src.backend import random_bernoulli
from tf_keras.src.backend import random_binomial
from tf_keras.src.backend import random_normal
from tf_keras.src.backend import random_normal_variable
from tf_keras.src.backend import random_uniform
from tf_keras.src.backend import random_uniform_variable
from tf_keras.src.backend import relu
from tf_keras.src.backend import repeat
from tf_keras.src.backend import repeat_elements
from tf_keras.src.backend import reset_uids
from tf_keras.src.backend import reshape
from tf_keras.src.backend import resize_images
from tf_keras.src.backend import resize_volumes
from tf_keras.src.backend import reverse
from tf_keras.src.backend import rnn
from tf_keras.src.backend import round
from tf_keras.src.backend import separable_conv2d
from tf_keras.src.backend import set_learning_phase
from tf_keras.src.backend import set_value
from tf_keras.src.backend import shape
from tf_keras.src.backend import sigmoid
from tf_keras.src.backend import sign
from tf_keras.src.backend import sin
from tf_keras.src.backend import softmax
from tf_keras.src.backend import softplus
from tf_keras.src.backend import softsign
from tf_keras.src.backend import sparse_categorical_crossentropy
from tf_keras.src.backend import spatial_2d_padding
from tf_keras.src.backend import spatial_3d_padding
from tf_keras.src.backend import sqrt
from tf_keras.src.backend import square
from tf_keras.src.backend import squeeze
from tf_keras.src.backend import stack
from tf_keras.src.backend import std
from tf_keras.src.backend import stop_gradient
from tf_keras.src.backend import sum
from tf_keras.src.backend import switch
from tf_keras.src.backend import tanh
from tf_keras.src.backend import temporal_padding
from tf_keras.src.backend import tile
from tf_keras.src.backend import to_dense
from tf_keras.src.backend import transpose
from tf_keras.src.backend import truncated_normal
from tf_keras.src.backend import update
from tf_keras.src.backend import update_add
from tf_keras.src.backend import update_sub
from tf_keras.src.backend import var
from tf_keras.src.backend import variable
from tf_keras.src.backend import zeros
from tf_keras.src.backend import zeros_like
from tf_keras.src.backend_config import epsilon
from tf_keras.src.backend_config import floatx
from tf_keras.src.backend_config import image_data_format
from tf_keras.src.backend_config import set_epsilon
from tf_keras.src.backend_config import set_floatx
from tf_keras.src.backend_config import set_image_data_format



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras 3D transposed convolution layer (sometimes called deconvolution)."""


import tensorflow.compat.v2 as tf

from tf_keras.src import activations
from tf_keras.src import constraints
from tf_keras.src import initializers
from tf_keras.src import regularizers
from tf_keras.src.dtensor import utils
from tf_keras.src.engine.base_layer import Layer
from tf_keras.src.engine.input_spec import InputSpec
from tf_keras.src.layers.convolutional.conv3d import Conv3D
from tf_keras.src.utils import conv_utils

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export(
    "keras.layers.Conv3DTranspose", "keras.layers.Convolution3DTranspose"
)
class Conv3DTranspose(Conv3D):
    """Transposed convolution layer (sometimes called Deconvolution).

    The need for transposed convolutions generally arises
    from the desire to use a transformation going in the opposite direction
    of a normal convolution, i.e., from something that has the shape of the
    output of some convolution to something that has the shape of its input
    while maintaining a connectivity pattern that is compatible with
    said convolution.

    When using this layer as the first layer in a model,
    provide the keyword argument `input_shape`
    (tuple of integers or `None`, does not include the sample axis),
    e.g. `input_shape=(128, 128, 128, 3)` for a 128x128x128 volume with 3
    channels if `data_format="channels_last"`.

    Args:
      filters: Integer, the dimensionality of the output space
        (i.e. the number of output filters in the convolution).
      kernel_size: An integer or tuple/list of 3 integers, specifying the
        depth, height and width of the 3D convolution window.
        Can be a single integer to specify the same value for
        all spatial dimensions.
      strides: An integer or tuple/list of 3 integers,
        specifying the strides of the convolution along the depth, height
          and width.
        Can be a single integer to specify the same value for
        all spatial dimensions.
        Specifying any stride value != 1 is incompatible with specifying
        any `dilation_rate` value != 1.
      padding: one of `"valid"` or `"same"` (case-insensitive).
        `"valid"` means no padding. `"same"` results in padding with zeros
        evenly to the left/right or up/down of the input such that output has
        the same height/width dimension as the input.
      output_padding: An integer or tuple/list of 3 integers,
        specifying the amount of padding along the depth, height, and
        width.
        Can be a single integer to specify the same value for all
        spatial dimensions.
        The amount of output padding along a given dimension must be
        lower than the stride along that same dimension.
        If set to `None` (default), the output shape is inferred.
      data_format: A string,
        one of `channels_last` (default) or `channels_first`.
        The ordering of the dimensions in the inputs.
        `channels_last` corresponds to inputs with shape
        `(batch_size, depth, height, width, channels)` while `channels_first`
        corresponds to inputs with shape
        `(batch_size, channels, depth, height, width)`.
        When unspecified, uses `image_data_format` value found in your Keras
        config file at `~/.keras/keras.json` (if exists) else 'channels_last'.
        Defaults to 'channels_last'.
      dilation_rate: an integer or tuple/list of 3 integers, specifying
        the dilation rate to use for dilated convolution.
        Can be a single integer to specify the same value for
        all spatial dimensions.
        Currently, specifying any `dilation_rate` value != 1 is
        incompatible with specifying any stride value != 1.
      activation: Activation function to use.
        If you don't specify anything, no activation is applied
        (see `keras.activations`).
      use_bias: Boolean, whether the layer uses a bias vector.
      kernel_initializer: Initializer for the `kernel` weights matrix
        (see `keras.initializers`). Defaults to 'glorot_uniform'.
      bias_initializer: Initializer for the bias vector
        (see `keras.initializers`). Defaults to 'zeros'.
      kernel_regularizer: Regularizer function applied to
        the `kernel` weights matrix
        (see `keras.regularizers`).
      bias_regularizer: Regularizer function applied to the bias vector
        (see `keras.regularizers`).
      activity_regularizer: Regularizer function applied to
        the output of the layer (its "activation")
        (see `keras.regularizers`).
      kernel_constraint: Constraint function applied to the kernel matrix
        (see `keras.constraints`).
      bias_constraint: Constraint function applied to the bias vector
        (see `keras.constraints`).

    Input shape:
      5D tensor with shape:
      `(batch_size, channels, depth, rows, cols)` if
      data_format='channels_first'
      or 5D tensor with shape:
      `(batch_size, depth, rows, cols, channels)` if
      data_format='channels_last'.

    Output shape:
      5D tensor with shape:
      `(batch_size, filters, new_depth, new_rows, new_cols)` if
        data_format='channels_first'
      or 5D tensor with shape:
      `(batch_size, new_depth, new_rows, new_cols, filters)` if
        data_format='channels_last'.
      `depth` and `rows` and `cols` values might have changed due to padding.
      If `output_padding` is specified::
      ```
      new_depth = ((depth - 1) * strides[0] + kernel_size[0] - 2 * padding[0] +
      output_padding[0])
      new_rows = ((rows - 1) * strides[1] + kernel_size[1] - 2 * padding[1] +
      output_padding[1])
      new_cols = ((cols - 1) * strides[2] + kernel_size[2] - 2 * padding[2] +
      output_padding[2])
      ```

    Returns:
      A tensor of rank 5 representing
      `activation(conv3dtranspose(inputs, kernel) + bias)`.

    Raises:
      ValueError: if `padding` is "causal".
      ValueError: when both `strides` > 1 and `dilation_rate` > 1.

    References:
      - [A guide to convolution arithmetic for deep
        learning](https://arxiv.org/abs/1603.07285v1)
      - [Deconvolutional
        Networks](https://www.matthewzeiler.com/mattzeiler/deconvolutionalnetworks.pdf)
    """

    @utils.allow_initializer_layout
    def __init__(
        self,
        filters,
        kernel_size,
        strides=(1, 1, 1),
        padding="valid",
        output_padding=None,
        data_format=None,
        dilation_rate=(1, 1, 1),
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs,
    ):
        super().__init__(
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            activation=activations.get(activation),
            use_bias=use_bias,
            kernel_initializer=initializers.get(kernel_initializer),
            bias_initializer=initializers.get(bias_initializer),
            kernel_regularizer=regularizers.get(kernel_regularizer),
            bias_regularizer=regularizers.get(bias_regularizer),
            activity_regularizer=regularizers.get(activity_regularizer),
            kernel_constraint=constraints.get(kernel_constraint),
            bias_constraint=constraints.get(bias_constraint),
            **kwargs,
        )

        self.output_padding = output_padding
        if self.output_padding is not None:
            self.output_padding = conv_utils.normalize_tuple(
                self.output_padding, 3, "output_padding", allow_zero=True
            )
            for stride, out_pad in zip(self.strides, self.output_padding):
                if out_pad >= stride:
                    raise ValueError(
                        "Strides must be greater than output padding. "
                        f"Received strides={self.strides}, "
                        f"output_padding={self.output_padding}."
                    )

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape)
        if len(input_shape) != 5:
            raise ValueError(
                "Inputs should have rank 5. "
                f"Received input_shape={input_shape}."
            )
        channel_axis = self._get_channel_axis()
        if input_shape.dims[channel_axis].value is None:
            raise ValueError(
                "The channel dimension of the inputs "
                "to `Conv3DTranspose` should be defined. "
                f"The input_shape received is {input_shape}, "
                f"where axis {channel_axis} (0-based) "
                "is the channel dimension, which found to be `None`."
            )
        input_dim = int(input_shape[channel_axis])
        kernel_shape = self.kernel_size + (self.filters, input_dim)
        self.input_spec = InputSpec(ndim=5, axes={channel_axis: input_dim})

        self.kernel = self.add_weight(
            "kernel",
            shape=kernel_shape,
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            trainable=True,
            dtype=self.dtype,
        )
        if self.use_bias:
            self.bias = self.add_weight(
                "bias",
                shape=(self.filters,),
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                trainable=True,
                dtype=self.dtype,
            )
        else:
            self.bias = None
        # Call Layer.build() to skip Conv.build() which we override here.
        Layer.build(self, input_shape)

    def call(self, inputs):
        inputs_shape = tf.shape(inputs)
        batch_size = inputs_shape[0]
        if self.data_format == "channels_first":
            d_axis, h_axis, w_axis = 2, 3, 4
        else:
            d_axis, h_axis, w_axis = 1, 2, 3

        depth = inputs_shape[d_axis]
        height = inputs_shape[h_axis]
        width = inputs_shape[w_axis]

        kernel_d, kernel_h, kernel_w = self.kernel_size
        stride_d, stride_h, stride_w = self.strides

        if self.output_padding is None:
            out_pad_d = out_pad_h = out_pad_w = None
        else:
            out_pad_d, out_pad_h, out_pad_w = self.output_padding

        # Infer the dynamic output shape:
        out_depth = conv_utils.deconv_output_length(
            depth,
            kernel_d,
            padding=self.padding,
            output_padding=out_pad_d,
            stride=stride_d,
        )
        out_height = conv_utils.deconv_output_length(
            height,
            kernel_h,
            padding=self.padding,
            output_padding=out_pad_h,
            stride=stride_h,
        )
        out_width = conv_utils.deconv_output_length(
            width,
            kernel_w,
            padding=self.padding,
            output_padding=out_pad_w,
            stride=stride_w,
        )
        if self.data_format == "channels_first":
            output_shape = (
                batch_size,
                self.filters,
                out_depth,
                out_height,
                out_width,
            )
            strides = (1, 1, stride_d, stride_h, stride_w)
        else:
            output_shape = (
                batch_size,
                out_depth,
                out_height,
                out_width,
                self.filters,
            )
            strides = (1, stride_d, stride_h, stride_w, 1)

        output_shape_tensor = tf.stack(output_shape)
        outputs = tf.nn.conv3d_transpose(
            inputs,
            self.kernel,
            output_shape_tensor,
            strides,
            data_format=conv_utils.convert_data_format(
                self.data_format, ndim=5
            ),
            padding=self.padding.upper(),
        )

        if not tf.executing_eagerly() and inputs.shape.rank:
            # Infer the static output shape:
            out_shape = self.compute_output_shape(inputs.shape)
            outputs.set_shape(out_shape)

        if self.use_bias:
            outputs = tf.nn.bias_add(
                outputs,
                self.bias,
                data_format=conv_utils.convert_data_format(
                    self.data_format, ndim=4
                ),
            )

        if self.activation is not None:
            return self.activation(outputs)
        return outputs

    def compute_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape).as_list()
        output_shape = list(input_shape)
        if self.data_format == "channels_first":
            c_axis, d_axis, h_axis, w_axis = 1, 2, 3, 4
        else:
            c_axis, d_axis, h_axis, w_axis = 4, 1, 2, 3

        kernel_d, kernel_h, kernel_w = self.kernel_size
        stride_d, stride_h, stride_w = self.strides

        if self.output_padding is None:
            out_pad_d = out_pad_h = out_pad_w = None
        else:
            out_pad_d, out_pad_h, out_pad_w = self.output_padding

        output_shape[c_axis] = self.filters
        output_shape[d_axis] = conv_utils.deconv_output_length(
            output_shape[d_axis],
            kernel_d,
            padding=self.padding,
            output_padding=out_pad_d,
            stride=stride_d,
        )
        output_shape[h_axis] = conv_utils.deconv_output_length(
            output_shape[h_axis],
            kernel_h,
            padding=self.padding,
            output_padding=out_pad_h,
            stride=stride_h,
        )
        output_shape[w_axis] = conv_utils.deconv_output_length(
            output_shape[w_axis],
            kernel_w,
            padding=self.padding,
            output_padding=out_pad_w,
            stride=stride_w,
        )
        return tf.TensorShape(output_shape)

    def get_config(self):
        config = super().get_config()
        config.pop("dilation_rate")
        config["output_padding"] = self.output_padding
        return config


# Alias

Convolution3DTranspose = Conv3DTranspose




# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Base class for N-D convolutional LSTM layers."""


import tensorflow.compat.v2 as tf

from tf_keras.src import activations
from tf_keras.src import backend
from tf_keras.src import constraints
from tf_keras.src import initializers
from tf_keras.src import regularizers
from tf_keras.src.engine import base_layer
from tf_keras.src.layers.rnn.base_conv_rnn import ConvRNN
from tf_keras.src.layers.rnn.dropout_rnn_cell_mixin import DropoutRNNCellMixin
from tf_keras.src.utils import conv_utils


class ConvLSTMCell(DropoutRNNCellMixin, base_layer.BaseRandomLayer):
    """Cell class for the ConvLSTM layer.

    Args:
      rank: Integer, rank of the convolution, e.g. "2" for 2D convolutions.
      filters: Integer, the dimensionality of the output space (i.e. the number
        of output filters in the convolution).
      kernel_size: An integer or tuple/list of n integers, specifying the
        dimensions of the convolution window.
      strides: An integer or tuple/list of n integers, specifying the strides of
        the convolution. Specifying any stride value != 1 is incompatible with
        specifying any `dilation_rate` value != 1.
      padding: One of `"valid"` or `"same"` (case-insensitive). `"valid"` means
        no padding. `"same"` results in padding evenly to the left/right or
        up/down of the input such that output has the same height/width
        dimension as the input.
      data_format: A string, one of `channels_last` (default) or
        `channels_first`. When unspecified, uses
        `image_data_format` value found in your TF-Keras config file at
        `~/.keras/keras.json` (if exists) else 'channels_last'.
        Defaults to 'channels_last'.
      dilation_rate: An integer or tuple/list of n integers, specifying the
        dilation rate to use for dilated convolution. Currently, specifying any
        `dilation_rate` value != 1 is incompatible with specifying any `strides`
        value != 1.
      activation: Activation function to use. If you don't specify anything, no
        activation is applied
        (ie. "linear" activation: `a(x) = x`).
      recurrent_activation: Activation function to use for the recurrent step.
      use_bias: Boolean, whether the layer uses a bias vector.
      kernel_initializer: Initializer for the `kernel` weights matrix, used for
        the linear transformation of the inputs.
      recurrent_initializer: Initializer for the `recurrent_kernel` weights
        matrix, used for the linear transformation of the recurrent state.
      bias_initializer: Initializer for the bias vector.
      unit_forget_bias: Boolean. If True, add 1 to the bias of the forget gate
      at initialization. Use in combination with `bias_initializer="zeros"`.
      This is recommended in [Jozefowicz et al., 2015](
      http://www.jmlr.org/proceedings/papers/v37/jozefowicz15.pdf)
      kernel_regularizer: Regularizer function applied to the `kernel` weights
        matrix.
      recurrent_regularizer: Regularizer function applied to the
        `recurrent_kernel` weights matrix.
      bias_regularizer: Regularizer function applied to the bias vector.
      kernel_constraint: Constraint function applied to the `kernel` weights
        matrix.
      recurrent_constraint: Constraint function applied to the
        `recurrent_kernel` weights matrix.
      bias_constraint: Constraint function applied to the bias vector.
      dropout: Float between 0 and 1. Fraction of the units to drop for the
        linear transformation of the inputs.
      recurrent_dropout: Float between 0 and 1. Fraction of the units to drop
        for the linear transformation of the recurrent state.
    Call arguments:
      inputs: A (2+ `rank`)D tensor.
      states:  List of state tensors corresponding to the previous timestep.
      training: Python boolean indicating whether the layer should behave in
        training mode or in inference mode. Only relevant when `dropout` or
        `recurrent_dropout` is used.
    """

    def __init__(
        self,
        rank,
        filters,
        kernel_size,
        strides=1,
        padding="valid",
        data_format=None,
        dilation_rate=1,
        activation="tanh",
        recurrent_activation="hard_sigmoid",
        use_bias=True,
        kernel_initializer="glorot_uniform",
        recurrent_initializer="orthogonal",
        bias_initializer="zeros",
        unit_forget_bias=True,
        kernel_regularizer=None,
        recurrent_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        recurrent_constraint=None,
        bias_constraint=None,
        dropout=0.0,
        recurrent_dropout=0.0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.rank = rank
        if self.rank > 3:
            raise ValueError(
                f"Rank {rank} convolutions are not currently "
                f"implemented. Received: rank={rank}"
            )
        self.filters = filters
        self.kernel_size = conv_utils.normalize_tuple(
            kernel_size, self.rank, "kernel_size"
        )
        self.strides = conv_utils.normalize_tuple(
            strides, self.rank, "strides", allow_zero=True
        )
        self.padding = conv_utils.normalize_padding(padding)
        self.data_format = conv_utils.normalize_data_format(data_format)
        self.dilation_rate = conv_utils.normalize_tuple(
            dilation_rate, self.rank, "dilation_rate"
        )
        self.activation = activations.get(activation)
        self.recurrent_activation = activations.get(recurrent_activation)
        self.use_bias = use_bias

        self.kernel_initializer = initializers.get(kernel_initializer)
        self.recurrent_initializer = initializers.get(recurrent_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.unit_forget_bias = unit_forget_bias

        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.recurrent_regularizer = regularizers.get(recurrent_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)

        self.kernel_constraint = constraints.get(kernel_constraint)
        self.recurrent_constraint = constraints.get(recurrent_constraint)
        self.bias_constraint = constraints.get(bias_constraint)

        self.dropout = min(1.0, max(0.0, dropout))
        self.recurrent_dropout = min(1.0, max(0.0, recurrent_dropout))
        self.state_size = (self.filters, self.filters)

    def build(self, input_shape):
        super().build(input_shape)
        if self.data_format == "channels_first":
            channel_axis = 1
        else:
            channel_axis = -1
        if input_shape[channel_axis] is None:
            raise ValueError(
                "The channel dimension of the inputs (last axis) should be "
                "defined. Found None. Full input shape received: "
                f"input_shape={input_shape}"
            )
        input_dim = input_shape[channel_axis]
        self.kernel_shape = self.kernel_size + (input_dim, self.filters * 4)
        recurrent_kernel_shape = self.kernel_size + (
            self.filters,
            self.filters * 4,
        )

        self.kernel = self.add_weight(
            shape=self.kernel_shape,
            initializer=self.kernel_initializer,
            name="kernel",
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
        )
        self.recurrent_kernel = self.add_weight(
            shape=recurrent_kernel_shape,
            initializer=self.recurrent_initializer,
            name="recurrent_kernel",
            regularizer=self.recurrent_regularizer,
            constraint=self.recurrent_constraint,
        )

        if self.use_bias:
            if self.unit_forget_bias:

                def bias_initializer(_, *args, **kwargs):
                    return backend.concatenate(
                        [
                            self.bias_initializer(
                                (self.filters,), *args, **kwargs
                            ),
                            initializers.get("ones")(
                                (self.filters,), *args, **kwargs
                            ),
                            self.bias_initializer(
                                (self.filters * 2,), *args, **kwargs
                            ),
                        ]
                    )

            else:
                bias_initializer = self.bias_initializer
            self.bias = self.add_weight(
                shape=(self.filters * 4,),
                name="bias",
                initializer=bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
            )
        else:
            self.bias = None

    def call(self, inputs, states, training=None):
        h_tm1 = states[0]  # previous memory state
        c_tm1 = states[1]  # previous carry state

        # dropout matrices for input units
        dp_mask = self.get_dropout_mask_for_cell(inputs, training, count=4)
        # dropout matrices for recurrent units
        rec_dp_mask = self.get_recurrent_dropout_mask_for_cell(
            h_tm1, training, count=4
        )

        if 0 < self.dropout < 1.0:
            inputs_i = inputs * dp_mask[0]
            inputs_f = inputs * dp_mask[1]
            inputs_c = inputs * dp_mask[2]
            inputs_o = inputs * dp_mask[3]
        else:
            inputs_i = inputs
            inputs_f = inputs
            inputs_c = inputs
            inputs_o = inputs

        if 0 < self.recurrent_dropout < 1.0:
            h_tm1_i = h_tm1 * rec_dp_mask[0]
            h_tm1_f = h_tm1 * rec_dp_mask[1]
            h_tm1_c = h_tm1 * rec_dp_mask[2]
            h_tm1_o = h_tm1 * rec_dp_mask[3]
        else:
            h_tm1_i = h_tm1
            h_tm1_f = h_tm1
            h_tm1_c = h_tm1
            h_tm1_o = h_tm1

        (kernel_i, kernel_f, kernel_c, kernel_o) = tf.split(
            self.kernel, 4, axis=self.rank + 1
        )
        (
            recurrent_kernel_i,
            recurrent_kernel_f,
            recurrent_kernel_c,
            recurrent_kernel_o,
        ) = tf.split(self.recurrent_kernel, 4, axis=self.rank + 1)

        if self.use_bias:
            bias_i, bias_f, bias_c, bias_o = tf.split(self.bias, 4)
        else:
            bias_i, bias_f, bias_c, bias_o = None, None, None, None

        x_i = self.input_conv(inputs_i, kernel_i, bias_i, padding=self.padding)
        x_f = self.input_conv(inputs_f, kernel_f, bias_f, padding=self.padding)
        x_c = self.input_conv(inputs_c, kernel_c, bias_c, padding=self.padding)
        x_o = self.input_conv(inputs_o, kernel_o, bias_o, padding=self.padding)
        h_i = self.recurrent_conv(h_tm1_i, recurrent_kernel_i)
        h_f = self.recurrent_conv(h_tm1_f, recurrent_kernel_f)
        h_c = self.recurrent_conv(h_tm1_c, recurrent_kernel_c)
        h_o = self.recurrent_conv(h_tm1_o, recurrent_kernel_o)

        i = self.recurrent_activation(x_i + h_i)
        f = self.recurrent_activation(x_f + h_f)
        c = f * c_tm1 + i * self.activation(x_c + h_c)
        o = self.recurrent_activation(x_o + h_o)
        h = o * self.activation(c)
        return h, [h, c]

    @property
    def _conv_func(self):
        if self.rank == 1:
            return backend.conv1d
        if self.rank == 2:
            return backend.conv2d
        if self.rank == 3:
            return backend.conv3d

    def input_conv(self, x, w, b=None, padding="valid"):
        conv_out = self._conv_func(
            x,
            w,
            strides=self.strides,
            padding=padding,
            data_format=self.data_format,
            dilation_rate=self.dilation_rate,
        )
        if b is not None:
            conv_out = backend.bias_add(
                conv_out, b, data_format=self.data_format
            )
        return conv_out

    def recurrent_conv(self, x, w):
        strides = conv_utils.normalize_tuple(
            1, self.rank, "strides", allow_zero=True
        )
        conv_out = self._conv_func(
            x, w, strides=strides, padding="same", data_format=self.data_format
        )
        return conv_out

    def get_config(self):
        config = {
            "filters": self.filters,
            "kernel_size": self.kernel_size,
            "strides": self.strides,
            "padding": self.padding,
            "data_format": self.data_format,
            "dilation_rate": self.dilation_rate,
            "activation": activations.serialize(self.activation),
            "recurrent_activation": activations.serialize(
                self.recurrent_activation
            ),
            "use_bias": self.use_bias,
            "kernel_initializer": initializers.serialize(
                self.kernel_initializer
            ),
            "recurrent_initializer": initializers.serialize(
                self.recurrent_initializer
            ),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "unit_forget_bias": self.unit_forget_bias,
            "kernel_regularizer": regularizers.serialize(
                self.kernel_regularizer
            ),
            "recurrent_regularizer": regularizers.serialize(
                self.recurrent_regularizer
            ),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "recurrent_constraint": constraints.serialize(
                self.recurrent_constraint
            ),
            "bias_constraint": constraints.serialize(self.bias_constraint),
            "dropout": self.dropout,
            "recurrent_dropout": self.recurrent_dropout,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


class ConvLSTM(ConvRNN):
    """Abstract N-D Convolutional LSTM layer (used as implementation base).

    Similar to an LSTM layer, but the input transformations
    and recurrent transformations are both convolutional.

    Args:
      rank: Integer, rank of the convolution, e.g. "2" for 2D convolutions.
      filters: Integer, the dimensionality of the output space
        (i.e. the number of output filters in the convolution).
      kernel_size: An integer or tuple/list of n integers, specifying the
        dimensions of the convolution window.
      strides: An integer or tuple/list of n integers,
        specifying the strides of the convolution.
        Specifying any stride value != 1 is incompatible with specifying
        any `dilation_rate` value != 1.
      padding: One of `"valid"` or `"same"` (case-insensitive).
        `"valid"` means no padding. `"same"` results in padding evenly to
        the left/right or up/down of the input such that output has the same
        height/width dimension as the input.
      data_format: A string,
        one of `channels_last` (default) or `channels_first`.
        The ordering of the dimensions in the inputs.
        `channels_last` corresponds to inputs with shape
        `(batch, time, ..., channels)`
        while `channels_first` corresponds to
        inputs with shape `(batch, time, channels, ...)`.
        When unspecified, uses
        `image_data_format` value found in your TF-Keras config file at
         `~/.keras/keras.json` (if exists) else 'channels_last'.
        Defaults to 'channels_last'.
      dilation_rate: An integer or tuple/list of n integers, specifying
        the dilation rate to use for dilated convolution.
        Currently, specifying any `dilation_rate` value != 1 is
        incompatible with specifying any `strides` value != 1.
      activation: Activation function to use.
        By default hyperbolic tangent activation function is applied
        (`tanh(x)`).
      recurrent_activation: Activation function to use
        for the recurrent step.
      use_bias: Boolean, whether the layer uses a bias vector.
      kernel_initializer: Initializer for the `kernel` weights matrix,
        used for the linear transformation of the inputs.
      recurrent_initializer: Initializer for the `recurrent_kernel`
        weights matrix,
        used for the linear transformation of the recurrent state.
      bias_initializer: Initializer for the bias vector.
      unit_forget_bias: Boolean.
        If True, add 1 to the bias of the forget gate at initialization.
        Use in combination with `bias_initializer="zeros"`.
        This is recommended in [Jozefowicz et al., 2015](
          http://www.jmlr.org/proceedings/papers/v37/jozefowicz15.pdf)
      kernel_regularizer: Regularizer function applied to
        the `kernel` weights matrix.
      recurrent_regularizer: Regularizer function applied to
        the `recurrent_kernel` weights matrix.
      bias_regularizer: Regularizer function applied to the bias vector.
      activity_regularizer: Regularizer function applied to.
      kernel_constraint: Constraint function applied to
        the `kernel` weights matrix.
      recurrent_constraint: Constraint function applied to
        the `recurrent_kernel` weights matrix.
      bias_constraint: Constraint function applied to the bias vector.
      return_sequences: Boolean. Whether to return the last output
        in the output sequence, or the full sequence. (default False)
      return_state: Boolean Whether to return the last state
        in addition to the output. (default False)
      go_backwards: Boolean (default False).
        If True, process the input sequence backwards.
      stateful: Boolean (default False). If True, the last state
        for each sample at index i in a batch will be used as initial
        state for the sample of index i in the following batch.
      dropout: Float between 0 and 1.
        Fraction of the units to drop for
        the linear transformation of the inputs.
      recurrent_dropout: Float between 0 and 1.
        Fraction of the units to drop for
        the linear transformation of the recurrent state.
    """

    def __init__(
        self,
        rank,
        filters,
        kernel_size,
        strides=1,
        padding="valid",
        data_format=None,
        dilation_rate=1,
        activation="tanh",
        recurrent_activation="hard_sigmoid",
        use_bias=True,
        kernel_initializer="glorot_uniform",
        recurrent_initializer="orthogonal",
        bias_initializer="zeros",
        unit_forget_bias=True,
        kernel_regularizer=None,
        recurrent_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        recurrent_constraint=None,
        bias_constraint=None,
        return_sequences=False,
        return_state=False,
        go_backwards=False,
        stateful=False,
        dropout=0.0,
        recurrent_dropout=0.0,
        **kwargs,
    ):
        cell = ConvLSTMCell(
            rank=rank,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            activation=activation,
            recurrent_activation=recurrent_activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            recurrent_initializer=recurrent_initializer,
            bias_initializer=bias_initializer,
            unit_forget_bias=unit_forget_bias,
            kernel_regularizer=kernel_regularizer,
            recurrent_regularizer=recurrent_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            recurrent_constraint=recurrent_constraint,
            bias_constraint=bias_constraint,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            name="conv_lstm_cell",
            dtype=kwargs.get("dtype"),
        )
        super().__init__(
            rank,
            cell,
            return_sequences=return_sequences,
            return_state=return_state,
            go_backwards=go_backwards,
            stateful=stateful,
            **kwargs,
        )
        self.activity_regularizer = regularizers.get(activity_regularizer)

    def call(self, inputs, mask=None, training=None, initial_state=None):
        return super().call(
            inputs, mask=mask, training=training, initial_state=initial_state
        )

    @property
    def filters(self):
        return self.cell.filters

    @property
    def kernel_size(self):
        return self.cell.kernel_size

    @property
    def strides(self):
        return self.cell.strides

    @property
    def padding(self):
        return self.cell.padding

    @property
    def data_format(self):
        return self.cell.data_format

    @property
    def dilation_rate(self):
        return self.cell.dilation_rate

    @property
    def activation(self):
        return self.cell.activation

    @property
    def recurrent_activation(self):
        return self.cell.recurrent_activation

    @property
    def use_bias(self):
        return self.cell.use_bias

    @property
    def kernel_initializer(self):
        return self.cell.kernel_initializer

    @property
    def recurrent_initializer(self):
        return self.cell.recurrent_initializer

    @property
    def bias_initializer(self):
        return self.cell.bias_initializer

    @property
    def unit_forget_bias(self):
        return self.cell.unit_forget_bias

    @property
    def kernel_regularizer(self):
        return self.cell.kernel_regularizer

    @property
    def recurrent_regularizer(self):
        return self.cell.recurrent_regularizer

    @property
    def bias_regularizer(self):
        return self.cell.bias_regularizer

    @property
    def kernel_constraint(self):
        return self.cell.kernel_constraint

    @property
    def recurrent_constraint(self):
        return self.cell.recurrent_constraint

    @property
    def bias_constraint(self):
        return self.cell.bias_constraint

    @property
    def dropout(self):
        return self.cell.dropout

    @property
    def recurrent_dropout(self):
        return self.cell.recurrent_dropout

    def get_config(self):
        config = {
            "filters": self.filters,
            "kernel_size": self.kernel_size,
            "strides": self.strides,
            "padding": self.padding,
            "data_format": self.data_format,
            "dilation_rate": self.dilation_rate,
            "activation": activations.serialize(self.activation),
            "recurrent_activation": activations.serialize(
                self.recurrent_activation
            ),
            "use_bias": self.use_bias,
            "kernel_initializer": initializers.serialize(
                self.kernel_initializer
            ),
            "recurrent_initializer": initializers.serialize(
                self.recurrent_initializer
            ),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "unit_forget_bias": self.unit_forget_bias,
            "kernel_regularizer": regularizers.serialize(
                self.kernel_regularizer
            ),
            "recurrent_regularizer": regularizers.serialize(
                self.recurrent_regularizer
            ),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
            "activity_regularizer": regularizers.serialize(
                self.activity_regularizer
            ),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "recurrent_constraint": constraints.serialize(
                self.recurrent_constraint
            ),
            "bias_constraint": constraints.serialize(self.bias_constraint),
            "dropout": self.dropout,
            "recurrent_dropout": self.recurrent_dropout,
        }
        base_config = super().get_config()
        del base_config["cell"]
        return dict(list(base_config.items()) + list(config.items()))

    @classmethod
    def from_config(cls, config):
        return cls(**config)




"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.regularizers import L1
from tf_keras.src.regularizers import L1 as l1
from tf_keras.src.regularizers import L1L2
from tf_keras.src.regularizers import L2
from tf_keras.src.regularizers import L2 as l2
from tf_keras.src.regularizers import OrthogonalRegularizer
from tf_keras.src.regularizers import OrthogonalRegularizer as orthogonal_regularizer
from tf_keras.src.regularizers import Regularizer
from tf_keras.src.regularizers import deserialize
from tf_keras.src.regularizers import get
from tf_keras.src.regularizers import l1_l2
from tf_keras.src.regularizers import serialize



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.applications.convnext import ConvNeXtBase
from tf_keras.src.applications.convnext import ConvNeXtLarge
from tf_keras.src.applications.convnext import ConvNeXtSmall
from tf_keras.src.applications.convnext import ConvNeXtTiny
from tf_keras.src.applications.convnext import ConvNeXtXLarge
from tf_keras.src.applications.convnext import decode_predictions
from tf_keras.src.applications.convnext import preprocess_input



"""Builds a vocabulary from inputs to the layer."""

import random
import string

import tensorflow as tf
from tensorflow.python.util.tf_export import keras_export

from tf_keras.src.layers import Layer


@keras_export("keras.layers.experimental.DynamicLookup")
class DynamicLookup(Layer):
    """A layer that builds a vocabulary from inputs.

    This layer maintains a vocabulary that is continuously updated based on the
    inputs passed in every forward pass. The frequency of the input is tracked
    and used to maintain a vocabulary. The very last index will be treated as
    the index. If `vocabulary_size=10`, OOV index will be 9.

    Args:
      vocabulary_size: Integer value representing size of the vocabulary to
        build.
      initial_vocabulary: The vocabulary to initialize the layer with. If a 1D
        tensor is provided, the vocabulary will be initialized with that tensor.
        If a `tf.DType` object is provided, a random tensor of that dtype and of
        length `vocabulary_size` will be generated as the initial vocabulary.
        Supported `tf.DType` values include `tf.int32`, `tf.int64` and
        `tf.string`.
      eviction_policy: The eviction policy for the vocabulary. Available options
        are string values like "LFU" (Least Frequently Used) and *more to come*.
        If not specified, the default eviction policy is "LFU". Expects a
        string.
      **kwargs: Arguments for super class.

    Attributes: get_vocabulary(): Returns a tensor representing the current
      vocabulary of the layer. If you want to look up the vocabulary keys given
      a set of indices, you can simply use `tf.gather(vocabulary, indices)`.

    Example:
    Here is an example to demonstrate how to use the DynamicLookup layer
    ```
      vocabulary_size = 3
      eviction_policy = "LFU"
      vocab = tf.constant(["apple", "banana", "cherry"])
      layer = DynamicLookup(
          vocabulary_size,
          vocab,
          eviction_policy=eviction_policy,
      )
      inputs = tf.constant([
          ["apple", "banana"],
      ])

      outputs = layer(inputs)
      tf.print(outputs)

      # you get the following output
      [[0 1]]

      # get top k vocab
      top_k_vocab = layer.get_top_vocabulary(2)
      tf.print(top_k_vocab)

      # you get the following output
      ["apple", "banana"]
      ```
      If you want to checkpoint the vocabulary or vocabulary frequency, see
      the following example

      ```
      checkpoint =
      tf.train.Checkpoint(vocabulary=self.vocabulary)
      checkpoint.write(filepath)
      ```
    """

    def __init__(
        self,
        vocabulary_size,
        initial_vocabulary,
        eviction_policy="LFU",
        **kwargs,
    ):
        """Initializes the DynamicLookup layer."""

        super().__init__(**kwargs)
        self.vocabulary_size = vocabulary_size
        self.eviction_policy = eviction_policy
        if tf.is_tensor(initial_vocabulary):
            self.initial_vocabulary = initial_vocabulary
        elif initial_vocabulary in (
            tf.string,
            tf.int32,
            tf.int64,
        ):
            self.initial_vocabulary = (
                DynamicLookup._generate_random_initial_vocab(
                    vocabulary_size, initial_vocabulary
                )
            )
        else:
            raise ValueError(
                "Either specify the initial vocabulary or provide a "
                "valid dtype. The dtype argument must be one of the "
                "following: tf.string, tf.int32, tf.int64."
            )
        # maintain a 20% bigger hash table
        self.internal_table_size = tf.cast(
            tf.floor(
                tf.multiply(
                    tf.cast(self.vocabulary_size, dtype=tf.float32), 1.2
                )
            ),
            dtype=tf.int32,
        )
        self.vocabulary_dtype = self.initial_vocabulary.dtype
        if self.eviction_policy == "LFU":
            self.vocabulary_table_keys = tf.Variable(
                initial_value=pad_tensor(
                    self.initial_vocabulary, self.internal_table_size
                ),
                shape=tf.TensorShape(self.internal_table_size),
                dtype=self.vocabulary_dtype,
                trainable=False,
                name="vocabulary_table_keys",
                per_worker_variable=True,
            )
            self.vocabulary_table_values = tf.Variable(
                initial_value=pad_tensor(
                    tf.zeros_like(self.initial_vocabulary, dtype=tf.int32),
                    self.internal_table_size,
                ),
                shape=tf.TensorShape(self.internal_table_size),
                dtype=tf.int32,
                trainable=False,
                name="vocabulary_table_values",
                per_worker_variable=True,
            )
        else:
            raise ValueError(
                "{} eviction policy is currently unsupported by DynamicLookup"
                " layer."
                " It currently only supports `LFU`".format(self.eviction_policy)
            )

        # TODO(b/268243335): add more eviction policy
        # TODO(b/268243996): provide multiple OOV

    def build(self, input_shape=None):
        self.vocabulary = self.add_weight(
            shape=self.initial_vocabulary.shape,
            dtype=self.vocabulary_dtype,
            initializer=tf.constant_initializer(
                self.initial_vocabulary.numpy()
            ),
            trainable=False,
            name="vocabulary",
        )
        super().build(input_shape)

    def call(self, inputs, learn_vocab=True):
        """Learn vocabulary from inputs and perform vocabulary lookup.

        Args:
          inputs: Input tensor, or dict/list/tuple of input tensors.
          learn_vocab: A boolean value that specifies whether the vocabulary
            should be learned from the layer inputs or not. Defaults to True.

        Returns:
          A tensor or list/tuple of tensors.
        """
        flattened_inputs = tf.reshape(inputs, [-1])
        # get unique values from inputs
        unique, _ = tf.unique(flattened_inputs)
        unique = tf.cast(unique, dtype=self.vocabulary_dtype)
        # learn vocab form inputs
        if learn_vocab and self.eviction_policy == "LFU":
            self.update_internal_vocabulary(unique)

        # lookup for inputs in self.vocabulary
        top_k_vocab = self.vocabulary
        lookup_values = tf.expand_dims(flattened_inputs, axis=-1)
        condition = tf.reduce_any(
            tf.equal(top_k_vocab, tf.expand_dims(lookup_values, -1)), axis=-1
        )
        # the very last index will be the OOV index
        indices = tf.where(
            condition,
            tf.argmax(
                tf.equal(top_k_vocab, tf.expand_dims(lookup_values, -1)),
                axis=-1,
            ),
            self.vocabulary_size,
        )
        # reshape output to the same shape as input
        out = tf.reshape(tf.squeeze(indices), tf.shape(inputs))
        return out

    def update_internal_vocabulary(self, unique):
        # get new keys
        unpadded_keys = remove_padding(self.vocabulary_table_keys)
        unpadded_values = remove_padding(self.vocabulary_table_values)
        table_expanded = tf.expand_dims(unpadded_keys, axis=0)
        unique_expanded = tf.expand_dims(unique, axis=0)
        new_keys = tf.sets.difference(
            unique_expanded, table_expanded, aminusb=True
        )
        number_of_new_keys = tf.shape(new_keys.values)[0]
        # get number of keys to be removed from vocab_frequency
        number_of_keys_to_remove = (
            tf.shape(unpadded_keys)[0]
            - self.internal_table_size
            + number_of_new_keys
        )
        number_of_keys_to_remove = tf.cast(number_of_keys_to_remove, tf.int32)
        number_of_keys_to_remove = tf.maximum(number_of_keys_to_remove, 0)
        # remove old keys
        updated_keys, updated_values = self._remove_old_keys(
            unpadded_keys,
            unpadded_values,
            number_of_keys_to_remove,
        )
        # add new keys
        self._add_new_keys(
            updated_keys,
            updated_values,
            unique,
            new_keys,
        )
        return unique

    def _remove_old_keys(self, unpadded_keys, unpadded_values, n):
        """remove old keys."""
        updated_keys, updated_values = None, None
        if self.eviction_policy == "LFU":
            # LFU eviction
            # negate the values of counts to find the lower n keys to remove
            negative_count = tf.math.negative(unpadded_values)
            # get index of lower n counts
            _, lower_n_index = tf.nn.top_k(negative_count, k=n)
            # gather keys that needs to be removed
            keys_to_remove = tf.gather(unpadded_keys, lower_n_index)
            # get masks for keys not present in inputs
            mask = tf.reduce_all(
                unpadded_keys[:, tf.newaxis] != keys_to_remove, axis=1
            )
            # updated keys and values with least frequent keys removed
            updated_keys = tf.boolean_mask(
                unpadded_keys,
                mask,
            )
            updated_values = tf.boolean_mask(
                unpadded_values,
                mask,
            )
        return updated_keys, updated_values

    def _add_new_keys(self, updated_keys, updated_values, unique, new_keys):
        """Add new keys and update internal vocabulary table."""
        if self.eviction_policy == "LFU":
            # increment values of old keys when present in current inputs
            matches = tf.where(
                tf.equal(tf.expand_dims(updated_keys, axis=1), unique)
            )[:, 0]
            updates = tf.ones_like(matches, dtype=tf.int32)
            matches = tf.expand_dims(matches, axis=-1)
            values_2 = tf.tensor_scatter_nd_add(
                updated_values, matches, updates
            )
            # add new keys and corresponding values = 1
            values_difference = tf.ones_like(new_keys.values, dtype=tf.int32)
            # concatenate old keys and new keys and pad
            updated_keys = pad_tensor(
                tf.concat([updated_keys, new_keys.values], axis=0),
                self.internal_table_size,
            )
            self.vocabulary_table_keys.assign(updated_keys)
            # concatenate updated old values and new values and pad
            updated_values = pad_tensor(
                tf.concat([values_2, values_difference], axis=0),
                self.internal_table_size,
            )
            self.vocabulary_table_values.assign(updated_values)
        return unique

    def get_top_vocabulary(self, k):
        """Get top k vocabulary keys."""
        top_k_vocab = None
        if self.eviction_policy == "LFU":
            values_len = tf.shape(self.vocabulary_table_keys)[0]
            if values_len > k:
                _, indices = tf.nn.top_k(self.vocabulary_table_values, k=k)
            else:
                _, indices = tf.nn.top_k(
                    self.vocabulary_table_values, k=values_len
                )
            top_k_vocab = tf.gather(self.vocabulary_table_keys, indices)
        return top_k_vocab

    def get_vocabulary(self):
        return self.vocabulary

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "vocabulary_size": self.vocabulary_size,
                "initial_vocabulary": self.initial_vocabulary.numpy().tolist(),
                "eviction_policy": self.eviction_policy,
            }
        )
        return config

    def save_assets(self, dir_path):
        vocabulary = self.vocabulary.numpy().tolist()
        vocabulary_filepath = tf.io.gfile.join(dir_path, "vocabulary.txt")
        with open(vocabulary_filepath, "w") as f:
            f.write("\n".join([str(w) for w in vocabulary]))

    def _generate_random_initial_vocab(
        vocabulary_size, dtype
    ):  # pylint: disable=no-self-argument
        if dtype == tf.string:
            chars = string.ascii_letters
            random_strings = [
                "".join([random.choice(chars) for _ in range(10)])
                for _ in range(vocabulary_size)
            ]
            random_vocab = tf.constant(random_strings, dtype=tf.string)
        elif dtype == tf.int32:
            random_vocab = tf.random.uniform(
                shape=[vocabulary_size],
                minval=0,
                maxval=vocabulary_size,
                dtype=tf.int32,
            )
        elif dtype == tf.int64:
            random_vocab = tf.random.uniform(
                shape=[vocabulary_size],
                minval=0,
                maxval=vocabulary_size,
                dtype=tf.int64,
            )
        else:
            raise ValueError(
                "Supported dtype for initial vocabulary include `tf.int32`,"
                " `tf.int64`, or `tf.string`. But got dtype = {}".format(dtype)
            )

        return random_vocab


def pad_tensor(tensor, n):
    """Pad a tensor to a fixed length."""
    if tensor.dtype == tf.string:
        padding = "unk"
    else:
        padding = -1
    pad_length = tf.maximum(n - tf.shape(tensor)[0], 0)
    padded_tensor = tf.pad(tensor, [[0, pad_length]], constant_values=padding)
    return padded_tensor[:n]


def remove_padding(tensor):
    """Remove padding from a tensor."""
    if tensor.dtype == tf.string:
        padding = "unk"
    else:
        padding = -1
    mask = tf.reshape(tensor != padding, shape=[-1])
    return tf.boolean_mask(tensor, mask)




"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.applications.resnet_v2 import ResNet101V2
from tf_keras.src.applications.resnet_v2 import ResNet152V2
from tf_keras.src.applications.resnet_v2 import ResNet50V2
from tf_keras.src.applications.resnet_v2 import decode_predictions
from tf_keras.src.applications.resnet_v2 import preprocess_input



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""Contains the base Layer class, from which all layers inherit."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
import warnings

import tensorflow.compat.v2 as tf

from tf_keras.src import backend
from tf_keras.src.engine import base_layer_utils
from tf_keras.src.engine import base_layer_v1 as base_layer
from tf_keras.src.legacy_tf_layers import variable_scope_shim
from tf_keras.src.mixed_precision import policy
from tf_keras.src.utils import tf_contextlib

# isort: off
from tensorflow.python.ops import variable_scope as vs
from tensorflow.python.util.tf_export import keras_export

_KERAS_STYLE_SCOPE = False


@keras_export(
    v1=["keras.__internal__.legacy.layers.experimental.keras_style_scope"]
)
@tf_contextlib.contextmanager
def keras_style_scope():
    """Use Keras-style variable management.

    All tf.layers and tf RNN cells created in this scope use Keras-style
    variable management.  Creating such layers with a scope= argument is
    disallowed, and reuse=True is disallowed.

    The purpose of this scope is to allow users of existing layers to
    slowly transition to a TF-Keras layers API without breaking existing
    functionality.

    One example of this is when using TensorFlow's RNN classes with Keras
    Models or Networks.  Because TF-Keras models do not properly set variable
    scopes, users of RNNs may either accidentally share scopes between two
    different models, or get errors about variables that already exist.

    Example:

    ```python
    class RNNModel(tf.keras.Model):

      def __init__(self, name):
        super(RNNModel, self).__init__(name=name)
        self.rnn = tf.compat.v1.nn.rnn_cell.MultiRNNCell(
          [tf.compat.v1.nn.rnn_cell.LSTMCell(64) for _ in range(2)])

      def call(self, input, state):
        return self.rnn(input, state)

    model_1 = RNNModel("model_1")
    model_2 = RNNModel("model_2")

    # OK
    output_1, next_state_1 = model_1(input, state)
    # Raises an error about trying to create an already existing variable.
    output_2, next_state_2 = model_2(input, state)
    ```

    The solution is to wrap the model construction and execution in a
    keras-style scope:

    ```python
    with keras_style_scope():
      model_1 = RNNModel("model_1")
      model_2 = RNNModel("model_2")

      # model_1 and model_2 are guaranteed to create their own variables.
      output_1, next_state_1 = model_1(input, state)
      output_2, next_state_2 = model_2(input, state)

      assert len(model_1.weights) > 0
      assert len(model_2.weights) > 0
      assert(model_1.weights != model_2.weights)
    ```

    Yields:
      A keras layer style scope.
    """
    global _KERAS_STYLE_SCOPE
    stack = _KERAS_STYLE_SCOPE
    _KERAS_STYLE_SCOPE = True
    try:
        yield
    finally:
        _KERAS_STYLE_SCOPE = stack


@keras_export(
    v1=["keras.__internal__.legacy.layers.experimental.set_keras_style"]
)
def set_keras_style():
    """Use Keras-style variable management.

    All tf.layers and tf RNN cells created after keras style ha been enabled
    use Keras-style variable management.  Creating such layers with a
    scope= argument is disallowed, and reuse=True is disallowed.

    The purpose of this function is to allow users of existing layers to
    slowly transition to TF-Keras layers API without breaking existing
    functionality.

    For more details, see the documentation for `keras_style_scope`.

    Note, once keras style has been set, it is set globally for the entire
    program and cannot be unset.

    Example:

    ```python
    set_keras_style()

    model_1 = RNNModel(name="model_1")
    model_2 = RNNModel(name="model_2")

    # model_1 and model_2 are guaranteed to create their own variables.
    output_1, next_state_1 = model_1(input, state)
    output_2, next_state_2 = model_2(input, state)

    assert len(model_1.weights) > 0
    assert len(model_2.weights) > 0
    assert(model_1.weights != model_2.weights)
    ```
    """
    global _KERAS_STYLE_SCOPE
    _KERAS_STYLE_SCOPE = True


def _is_in_keras_style_scope():
    global _KERAS_STYLE_SCOPE
    return _KERAS_STYLE_SCOPE


@keras_export(v1=["keras.__internal__.legacy.layers.Layer"])
class Layer(base_layer.Layer):
    """Base layer class.

    It is considered legacy, and we recommend the use of `tf.keras.layers.Layer`
    instead.

    Args:
      trainable: Boolean, whether the layer's variables should be trainable.
      name: String name of the layer.
      dtype: Default dtype of the layer's weights (default of `None` means use
        the type of the first input).

    Read-only properties:
      name: The name of the layer (string).
      dtype: Default dtype of the layer's weights (default of `None` means use
        the type of the first input).
      trainable_variables: List of trainable variables.
      non_trainable_variables: List of non-trainable variables.
      variables: List of all variables of this layer, trainable and
        non-trainable.
      updates: List of update ops of this layer.
      losses: List of losses added by this layer.
      trainable_weights: List of variables to be included in backprop.
      non_trainable_weights: List of variables that should not be
        included in backprop.
      weights: The concatenation of the lists trainable_weights and
        non_trainable_weights (in this order).

    Mutable properties:
      trainable: Whether the layer should be trained (boolean).
      input_spec: Optional (list of) `InputSpec` object(s) specifying the
        constraints on inputs that can be accepted by the layer.
    """

    def __init__(self, trainable=True, name=None, dtype=None, **kwargs):
        # For backwards compatibility, legacy layers do not use
        # `ResourceVariable` by default.
        self._use_resource_variables = False
        scope = kwargs.pop("_scope", None)
        self._reuse = kwargs.pop("_reuse", None)

        # Avoid an incorrect lint error
        self._trainable_weights = []
        self.built = False

        if dtype is None:
            # Indicates to infer dtype from inputs. When the V2 dtype behavior
            # is enabled, TF-Keras layers default their dtype to floatx instead,
            # so we pass an "_infer" policy to keep the old V1 behavior.
            dtype = policy.Policy("_infer")

        if "autocast" not in kwargs:
            kwargs["autocast"] = False

        # Mark that legacy layers should not be instrumented as TF-Keras usage
        self._disable_keras_instrumentation = True

        super().__init__(trainable=trainable, name=name, dtype=dtype, **kwargs)

        if _is_in_keras_style_scope():
            if scope is not None:
                raise ValueError(
                    "scope argument not allowed when keras style layers are "
                    "enabled, but saw: {}".format(scope)
                )
            if self._reuse is not None:
                raise ValueError(
                    "reuse argument not allowed when keras style layers are "
                    "enabled, but saw: {}".format(self._reuse)
                )
            self._keras_style = True
        else:
            self._keras_style = False

        self._call_has_scope_arg = "scope" in self._call_spec.arg_names
        if scope:
            with tf.compat.v1.variable_scope(scope) as captured_scope:
                self._scope = captured_scope
        else:
            self._scope = None
        self._current_scope = None

    def apply(self, *args, **kwargs):
        return self(*args, **kwargs)

    # We no longer track graph in tf.layers layers. This property is only kept
    # to maintain API backward compatibility.
    @property
    def graph(self):
        warnings.warn(
            "`Layer.graph` is deprecated and "
            "will be removed in a future version. "
            "Please stop using this property because tf.layers layers no "
            "longer track their graph.",
            stacklevel=2,
        )
        if tf.executing_eagerly():
            raise RuntimeError(
                "Layer.graph not supported when executing eagerly."
            )
        return None

    def _init_set_name(self, name):
        # Determine layer name (non-unique).
        if isinstance(name, tf.compat.v1.VariableScope):
            base_name = name.name
            self._name, _ = self._make_unique_name()
        else:
            base_name = name
            self._name = name
        if not name:
            self._name, base_name = self._make_unique_name()
        self._base_name = base_name

    def _make_unique_name(
        self,
        name_uid_map=None,
        avoid_names=None,
        namespace="",
        zero_based=False,
    ):
        base_name = base_layer.to_snake_case(self.__class__.__name__)
        name = backend.unique_object_name(
            base_name,
            name_uid_map=name_uid_map,
            avoid_names=avoid_names,
            namespace=namespace,
            zero_based=zero_based,
        )
        return (name, base_name)

    @property
    def scope_name(self):
        if not self._scope:
            raise ValueError(
                'No name available for layer scope because the layer "'
                + self._name
                + '" has not been used yet. The scope name '
                + " is determined the first time the layer instance is "
                + "called. You must therefore call the layer before "
                + "querying `scope_name`."
            )
        return self._scope.name

    def add_loss(self, losses, inputs=None):
        previous_losses_length = len(self._losses)
        previous_callable_losses_length = len(self._callable_losses)
        super().add_loss(losses, inputs=inputs)
        if not tf.executing_eagerly():
            # TODO(fchollet): deprecate collection below.
            new_losses = self._losses[previous_losses_length:]
            new_callable_losses = self._callable_losses[
                previous_callable_losses_length:
            ]
            for regularizer in new_callable_losses:
                loss_tensor = regularizer()
                if loss_tensor is not None:
                    new_losses.append(loss_tensor)
            _add_elements_to_collection(
                new_losses, tf.compat.v1.GraphKeys.REGULARIZATION_LOSSES
            )

    def _name_scope(self):
        """Determines op naming for the Layer."""
        if self._keras_style:
            return super()._name_scope()
        return self._current_scope.original_name_scope

    def _set_scope(self, scope=None):
        if self._scope is None:
            # If constructed with _scope=None, lazy setting of scope.
            if self._reuse:
                with tf.compat.v1.variable_scope(
                    scope if scope is not None else self._base_name
                ) as captured_scope:
                    self._scope = captured_scope
            else:
                with tf.compat.v1.variable_scope(
                    scope, default_name=self._base_name
                ) as captured_scope:
                    self._scope = captured_scope

    def add_weight(
        self,
        name,
        shape,
        dtype=None,
        initializer=None,
        regularizer=None,
        trainable=None,
        constraint=None,
        use_resource=None,
        synchronization=tf.VariableSynchronization.AUTO,
        aggregation=tf.compat.v1.VariableAggregation.NONE,
        partitioner=None,
        **kwargs
    ):
        """Adds a new variable to the layer, or gets an existing one; returns it

        Args:
          name: variable name.
          shape: variable shape.
          dtype: The type of the variable. Defaults to `self.dtype` or
            `float32`.
          initializer: initializer instance (callable).
          regularizer: regularizer instance (callable).
          trainable: whether the variable should be part of the layer's
            "trainable_variables" (e.g. variables, biases)
            or "non_trainable_variables" (e.g. BatchNorm mean, stddev).
            Note, if the current variable scope is marked as non-trainable
            then this parameter is ignored and any added variables are also
            marked as non-trainable. `trainable` becomes `True` unless
            `synchronization` is set to `ON_READ`. Defaults to `True`.
          constraint: constraint instance (callable).
          use_resource: Whether to use `ResourceVariable`.
          synchronization: Indicates when a distributed a variable will be
            aggregated. Accepted values are constants defined in the class
            `tf.VariableSynchronization`. By default the synchronization is set
            to `AUTO` and the current `DistributionStrategy` chooses when to
            synchronize. If `synchronization` is set to `ON_READ`, `trainable`
            must not be set to `True`.
          aggregation: Indicates how a distributed variable will be aggregated.
            Accepted values are constants defined in the class
            `tf.VariableAggregation`.
          partitioner: (optional) partitioner instance (callable).  If
            provided, when the requested variable is created it will be split
            into multiple partitions according to `partitioner`.  In this case,
            an instance of `PartitionedVariable` is returned.  Available
            partitioners include `tf.compat.v1.fixed_size_partitioner` and
            `tf.compat.v1.variable_axis_size_partitioner`.  For more details,
            see the documentation of `tf.compat.v1.get_variable` and the
            "Variable Partitioners and Sharding" section of the API guide.
          **kwargs: Additional keyword arguments.

        Returns:
          The created variable.  Usually either a `Variable` or
          `ResourceVariable` instance.  If `partitioner` is not `None`, a
          `PartitionedVariable` instance is returned.

        Raises:
          RuntimeError: If called with partitioned variable regularization and
            eager execution is enabled.
          ValueError: When trainable has been set to True with synchronization
            set as `ON_READ`.
        """
        for kwarg in kwargs:
            if kwarg != "experimental_autocast":
                raise TypeError("Unknown keyword argument:", kwarg)
        if self._keras_style:
            return super().add_weight(
                name=name,
                shape=shape,
                dtype=dtype,
                initializer=initializer,
                regularizer=regularizer,
                trainable=trainable and self.trainable,
                constraint=constraint,
                use_resource=use_resource,
                synchronization=tf.VariableSynchronization.AUTO,
                aggregation=tf.compat.v1.VariableAggregation.NONE,
                partitioner=partitioner,
                **kwargs
            )

        if synchronization == tf.VariableSynchronization.ON_READ:
            if trainable:
                raise ValueError(
                    "Synchronization value can be set to "
                    "VariableSynchronization.ON_READ only for non-trainable "
                    "variables. You have specified trainable=True and "
                    "synchronization=VariableSynchronization.ON_READ."
                )
            else:
                # Set trainable to be false when variable is to be synced on
                # read.
                trainable = False
        elif trainable is None:
            trainable = True

        def _should_add_regularizer(variable, existing_variable_set):
            if base_layer_utils.is_split_variable(variable):
                for var in variable:
                    if var in existing_variable_set:
                        return False
                return True
            else:
                return variable not in existing_variable_set

        init_graph = None
        if not tf.executing_eagerly():
            default_graph = tf.compat.v1.get_default_graph()
            if default_graph.building_function:
                with tf.init_scope():
                    # Retrieve the variables from the graph into which variables
                    # will be lifted; if initialization ops will be lifted into
                    # the eager context, then there is nothing to retrieve,
                    # since variable collections are not supported when eager
                    # execution is enabled.
                    if not tf.executing_eagerly():
                        init_graph = tf.compat.v1.get_default_graph()
                        existing_variables = set(
                            tf.compat.v1.global_variables()
                        )
            else:
                # Initialization ops will not be lifted out of the default
                # graph.
                init_graph = default_graph
                existing_variables = set(tf.compat.v1.global_variables())

        if dtype is None:
            dtype = self.dtype or tf.float32

        self._set_scope(None)
        reuse = self.built or self._reuse
        prev_len_trainable = len(self._trainable_weights)
        with tf.compat.v1.variable_scope(
            self._scope, reuse=reuse, auxiliary_name_scope=False
        ) as scope:
            self._current_scope = scope
            with backend.name_scope(self._name_scope()):
                use_resource = (
                    use_resource
                    or self._use_resource_variables
                    or scope.use_resource
                )
                if initializer is None:
                    initializer = scope.initializer
                variable = super().add_weight(
                    name,
                    shape,
                    dtype=tf.as_dtype(dtype),
                    initializer=initializer,
                    trainable=trainable and self.trainable,
                    constraint=constraint,
                    partitioner=partitioner,
                    use_resource=use_resource,
                    synchronization=synchronization,
                    aggregation=aggregation,
                    getter=tf.compat.v1.get_variable,
                    **kwargs
                )

                if regularizer:
                    if (
                        tf.compat.v1.executing_eagerly_outside_functions()
                        or _should_add_regularizer(variable, existing_variables)
                    ):
                        self._handle_weight_regularization(
                            name, variable, regularizer
                        )
                        var_store = vs._get_default_variable_store()
                        # When the shim to get variable scope working in TF2 is
                        # used, We need to explicitly make the shim track the
                        # regularization losses as the collections will not be
                        # accessible.
                        if hasattr(var_store, "add_regularizer"):
                            var_store.add_regularizer(variable, regularizer)

                if init_graph is not None:
                    # Handle edge case where a custom getter has overridden
                    # `trainable`.  There is one known occurrence of this, in
                    # unit test testBasicRNNCellNotTrainable in
                    # contrib.rnn.python.kernel_tests.core_rnn_cell_test
                    with init_graph.as_default():
                        trainable_variables = tf.compat.v1.trainable_variables()
                    if (
                        trainable
                        and self.trainable
                        and variable not in trainable_variables
                    ):
                        # A custom getter / variable scope overrode the
                        # trainable flag.
                        extra_trainable_vars = self._trainable_weights[
                            prev_len_trainable:
                        ]
                        self._trainable_weights = self._trainable_weights[
                            :prev_len_trainable
                        ]
                        self._non_trainable_weights += extra_trainable_vars
        return variable

    def __call__(self, inputs, *args, **kwargs):
        """Wraps `call`, applying pre- and post-processing steps.

        Args:
          inputs: input tensor(s).
          *args: additional positional arguments to be passed to `self.call`.
          **kwargs: additional keyword arguments to be passed to `self.call`.
            **Note**: kwarg `scope` is reserved for use by the layer.

        Returns:
          Output tensor(s).

        Note:
          - If the layer's `call` method takes a `scope` keyword argument, this
            argument will be automatically set to the current variable scope.
          - If the layer's `call` method takes a `mask` argument (as some Keras
            layers do), its default value will be set to the mask generated
            for `inputs` by the previous layer (if `input` did come from
            a layer that generated a corresponding mask, i.e. if it came from
            a TF-Keras layer with masking support.

        Raises:
          ValueError: if the layer's `call` method returns None (an invalid
            value).
        """
        scope = kwargs.pop("scope", None)

        if self._keras_style:
            if scope is not None:
                raise ValueError(
                    "scope argument not allowed when keras style layers are "
                    "enabled, but saw: {}".format(scope)
                )
            return super().__call__(inputs, *args, **kwargs)

        self._set_scope(scope)

        if self.built:
            try:
                # Some classes which inherit from Layer do not use its
                # constructor, so rather than initializing to None we check for
                # an AttributeError.
                scope_context_manager = self._always_reuse_variable_scope
            except AttributeError:
                scope_context_manager = None

            if scope_context_manager is None:
                # From this point we will always set reuse=True, so create a
                # "final" variable scope with this setting. We avoid re-creating
                # variable scopes after this point as an optimization.
                scope_context_manager = tf.compat.v1.variable_scope(
                    self._scope, reuse=True, auxiliary_name_scope=False
                )

                # Do not cache variable scopes if Eager mode is enabled. If
                # Eager mode is enabled then we don't want to reuse scopes
                # because the cached scope might be from a FuncGraph or Eager
                # scope we are no longer in.
                if not tf.compat.v1.executing_eagerly_outside_functions():
                    self._always_reuse_variable_scope = scope_context_manager
        else:
            scope_context_manager = tf.compat.v1.variable_scope(
                self._scope, reuse=self._reuse, auxiliary_name_scope=False
            )

        with scope_context_manager as scope:
            self._current_scope = scope

            try:
                call_has_scope_arg = self._call_has_scope_arg
            except AttributeError:
                self._call_spec.arg_names = variable_scope_shim.fn_args(
                    self.call
                )
                self._call_has_scope_arg = "scope" in self._call_spec.arg_names
                call_has_scope_arg = self._call_has_scope_arg
            if call_has_scope_arg:
                kwargs["scope"] = scope

            # Actually call layer
            outputs = super().__call__(inputs, *args, **kwargs)

        if not tf.executing_eagerly():
            # Update global default collections.
            _add_elements_to_collection(
                self.updates, tf.compat.v1.GraphKeys.UPDATE_OPS
            )
        return outputs

    def __deepcopy__(self, memo):
        no_copy = set(["_graph", "_thread_local", "_metrics_lock"])
        shallow_copy = set(["_scope", "_always_reuse_variable_scope"])
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k in no_copy:
                setattr(result, k, v)
            elif k in shallow_copy:
                setattr(result, k, copy.copy(v))
            elif base_layer.is_tensor_or_tensor_list(v):
                setattr(result, k, v)
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        return result

    def __setattr__(self, value, name):
        # By-pass the automatic dependency tracking performed by the parent
        # Layer.
        super(tf.__internal__.tracking.Trackable, self).__setattr__(value, name)

    @property
    def _is_legacy_layer(self):
        """Used by keras to check compatibility. This should not be
        overridden."""
        return True


def _add_elements_to_collection(elements, collection_list):
    if tf.executing_eagerly():
        raise RuntimeError(
            "Using collections from Layers not supported in Eager "
            "mode. Tried to add %s to %s" % (elements, collection_list)
        )
    elements = tf.nest.flatten(elements)
    collection_list = tf.nest.flatten(collection_list)
    for name in collection_list:
        collection = tf.compat.v1.get_collection_ref(name)
        collection_set = {id(e) for e in collection}
        for element in elements:
            if id(element) not in collection_set:
                collection.append(element)




"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.api._v1.keras.callbacks import experimental
from tf_keras.src.callbacks import BaseLogger
from tf_keras.src.callbacks import CSVLogger
from tf_keras.src.callbacks import Callback
from tf_keras.src.callbacks import CallbackList
from tf_keras.src.callbacks import EarlyStopping
from tf_keras.src.callbacks import History
from tf_keras.src.callbacks import LambdaCallback
from tf_keras.src.callbacks import LearningRateScheduler
from tf_keras.src.callbacks import ModelCheckpoint
from tf_keras.src.callbacks import ProgbarLogger
from tf_keras.src.callbacks import ReduceLROnPlateau
from tf_keras.src.callbacks import RemoteMonitor
from tf_keras.src.callbacks import TerminateOnNaN
from tf_keras.src.callbacks_v1 import TensorBoard
from tf_keras.src.utils.sidecar_evaluator import SidecarEvaluatorModelExport



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.optimizers.adadelta import Adadelta
from tf_keras.src.optimizers.adagrad import Adagrad
from tf_keras.src.optimizers.adam import Adam
from tf_keras.src.optimizers.adamw import AdamW
from tf_keras.src.optimizers.rmsprop import RMSprop
from tf_keras.src.optimizers.sgd import SGD



"""A layer that updates its vocab and embedding matrix during training."""

import tensorflow as tf
from absl import logging
from tensorflow.python.util.tf_export import keras_export

import tf_keras.src as tf_keras
from tf_keras.src.layers import Layer
from tf_keras.src.layers.experimental import dynamic_lookup
from tf_keras.src.utils import warmstart_embedding_matrix


@keras_export("keras.layers.experimental.DynamicEmbedding")
class DynamicEmbedding(Layer):
    """A layer that updates its vocab and embedding matrix during training.

      DynamicEmbedding allows for the continuous updating of the vocabulary
      and embeddings during the training process. In traditional methods, the
      vocabulary and mapping to the embedding vector are set at the beginning of
      the training process and remain fixed throughout the training process.
      However, in many real-world scenarios, the vocabulary and mapping to the
      embeddings need to be updated to reflect the changing nature of the data.

      For instance, in natural language processing tasks, the vocabulary of
      words in a corpus may change over time, and it's important to update the
      embeddings to reflect these changes. Similarly, in recommendation systems,
      the items in the vocabulary may change over time.

      A layer that supports dynamic embeddings addresses this issue by allowing
      for the continuous updating of the vocabulary and embeddings during the
      training process. and it also updates the embedding matrix to reflect the
      new vocabulary.

      This layer maintains a hash table to track the most up-to-date vocabulary
      based on the inputs received by the layer and the eviction policy. When
      this layer is used with an `UpdateEmbeddingCallback`, which is a
      time-based callback, the vocabulary lookup tensor is updated at the time
      interval set in the `UpdateEmbeddingCallback` based on the most up-to-date
      vocabulary hash table maintained by the layer. If this layer is not used
      in conjunction with `UpdateEmbeddingCallback` the behavior of the layer
      would be same as `keras.layers.Embedding`.

    Args:
      input_dim: Size of the vocabulary in the input data. Expects an integer.
      output_dim: The size of the embedding space. Expects an integer.
      initial_vocabulary: The vocabulary to initialize the layer with. If a 1D
        tensor is provided, the vocabulary will be initialized with that tensor.
        If a `tf.DType` object is provided, a random tensor of that dtype and of
        length `input_dim` will be generated as the initial vocabulary.
        Supported `tf.DType` values include `tf.int32`, `tf.int64` and
        `tf.string`.
      eviction_policy: The eviction policy for the vocabulary. Available options
        are "LFU" (Least Frequently Used) and *more to come*. Defaults to "LFU".
        Expects a string.
      input_length:  Length of input sequences, when it is constant. This
        argument is required if you are going to connect `Flatten` then `Dense`
        layers upstream (without it, the shape of the dense outputs cannot be
        computed).Expects an integer.
      embedding_initializer: Initializer for embedding vectors for new input
        vocab tokens to be added to the updated embedding matrix (see
        keras.initializers). Defaults to "uniform".
      num_oov_indices: Number of out of vocabulary token to use. Currently
        supports 1. Expects an integer.
      **kwargs: Additional keyword arguments for the parent class.

    Attributes:
      embedding_layer: Embedding layer of DynamicEmbedding layer.
      dynamic_lookup_layer: DynamicLookup layer of DynamicEmbedding layer.
      embedding_initializer: Initializer for embedding vectors for new input
        vocab tokens to be added to the updated embedding matrix (see
        keras.initializers).
      num_oov_indices: Number of out of vocabulary token to use.

    Example:
    ```
      # Generate dummy data
      train_data = np.array([
          ['a', 'j', 'c', 'd', 'e'],
          ['a', 'h', 'i', 'j', 'b'],
          ['i', 'h', 'c', 'j', 'e'],
      ])
      train_labels = np.array([0, 1, 2])
      vocab = tf.constant(['a', 'b', 'c', 'd', 'e'])
      eviction_policy = 'LFU'
      # Define the model
      model = tf.keras.models.Sequential([
          DynamicEmbedding(
              input_dim=5,
              output_dim=2,
              input_length=5,
              eviction_policy=eviction_policy,
              initial_vocabulary=vocab,
          ),
          tf.keras.layers.Flatten(),
          tf.keras.layers.Dense(3, activation='softmax'),
      ])

      # Compile the model
      model.compile(
          optimizer='adam',
          loss='sparse_categorical_crossentropy',
          metrics=['accuracy'],
      )
      # update the vocabulary every 1 second
      update_embedding_callback = UpdateEmbeddingCallback(
          model.layers[0], interval=1
      )
      with update_embedding_callback:
        result = model.fit(
            train_data,
            train_labels,
            epochs=100,
            batch_size=1,
            callbacks=[update_embedding_callback],
        )
    ```
    """

    def __init__(
        self,
        input_dim,
        output_dim,
        initial_vocabulary,
        eviction_policy="LFU",
        input_length=None,
        embedding_initializer="uniform",
        num_oov_indices=1,
        **kwargs,
    ):
        """Initialize DynamicEmbedding layer."""

        super().__init__(**kwargs)
        # assuming one oov bucket for now
        self.embedding_layer = tf_keras.layers.Embedding(
            input_dim=input_dim + num_oov_indices,
            output_dim=output_dim,
            embeddings_initializer=embedding_initializer,
            input_length=input_length,
            **kwargs,
        )
        self.dynamic_lookup_layer = dynamic_lookup.DynamicLookup(
            vocabulary_size=input_dim,
            eviction_policy=eviction_policy,
            initial_vocabulary=initial_vocabulary,
            **kwargs,
        )
        self.embedding_initializer = embedding_initializer
        self.num_oov_indices = num_oov_indices

    def build(self, input_shape=None):
        self.embedding_layer.build(input_shape)
        self.dynamic_lookup_layer.build(input_shape)

    def call(self, inputs, learn_vocab=True):
        # get vocab to index mapped for dynamic_lookup_layer
        output = self.dynamic_lookup_layer(inputs, learn_vocab=learn_vocab)
        # pass the indices as inputs to embedding_layer
        return self.embedding_layer(output)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "input_dim": self.embedding_layer.input_dim,
                "output_dim": self.embedding_layer.output_dim,
                "input_length": self.embedding_layer.input_length,
                "eviction_policy": self.dynamic_lookup_layer.eviction_policy,
                "initial_vocabulary": (
                    self.dynamic_lookup_layer.initial_vocabulary.numpy().tolist()  # noqa E501
                ),
                "embedding_initializer": self.embedding_initializer,
                "num_oov_indices": self.num_oov_indices,
            }
        )
        return config

    def get_vocabulary(self):
        return self.dynamic_lookup_layer.get_vocabulary()

    def save_assets(self, dir_path):
        initial_vocabulary = (
            self.dynamic_lookup_layer.initial_vocabulary.numpy().tolist()
        )
        initial_vocabulary_filepath = tf.io.gfile.join(
            dir_path, "initial_vocabulary.txt"
        )
        with open(initial_vocabulary_filepath, "w") as f:
            f.write("\n".join([str(w) for w in initial_vocabulary]))

    def update_embeddings(self, strategy):
        """Update embedding matrix of dynamic embedding layer."""
        try:
            if isinstance(strategy, tf.distribute.ParameterServerStrategy):
                # if using PSS agrregate values
                keys_list = (
                    self.dynamic_lookup_layer.vocabulary_table_keys.read_all()
                )
                values_list = (
                    self.dynamic_lookup_layer.vocabulary_table_values.read_all()
                )
                keys, values = self.aggregate_lookup_table(
                    keys_list, values_list
                )
            else:
                # if using on device strategy, just read values
                keys, values = (
                    self.dynamic_lookup_layer.vocabulary_table_keys,
                    self.dynamic_lookup_layer.vocabulary_table_values,
                )
            old_vocab = self.dynamic_lookup_layer.vocabulary
            new_vocab = self.get_top_vocabulary(
                keys,
                values,
                self.dynamic_lookup_layer.vocabulary_size,
            )
            # remap and update the embedding matrix
            embedding_matrix = self.embedding_layer.embeddings
            oov_token = tf.fill([self.num_oov_indices], "UNK")
            updated_new_vocab = tf.concat([new_vocab, oov_token], axis=0)
            embedding_matrix = warmstart_embedding_matrix(
                base_vocabulary=list(old_vocab.numpy()),
                new_vocabulary=updated_new_vocab,
                base_embeddings=embedding_matrix,
                new_embeddings_initializer=self.embedding_initializer,
            )
            self.dynamic_lookup_layer.vocabulary.assign(new_vocab)
            self.embedding_layer.embeddings.assign(embedding_matrix)
        except AttributeError:
            logging.info(
                "Time interval specified by the UpdateEmbeddingCallback may be"
                " too small, please try increasing the value of `interval`."
            )

    def aggregate_lookup_table(self, keys_list, values_list):
        # Flatten the keys and values matrices
        keys_1d = tf.reshape(keys_list, [-1])
        values_1d = tf.reshape(values_list, [-1])
        # Get unique keys and their corresponding summed values
        unique_keys, idx, _ = tf.unique_with_counts(keys_1d)
        summed_values = tf.math.unsorted_segment_sum(
            values_1d, idx, tf.shape(unique_keys)[0]
        )
        return unique_keys, summed_values

    def get_top_vocabulary(self, keys, values, k):
        """Get Top vocabulary keys and values."""
        values_len = tf.shape(keys)[0]
        if values_len > k:
            _, indices = tf.nn.top_k(values, k=k)
        else:
            _, indices = tf.nn.top_k(values, k=values_len)
        top_k_vocab = tf.gather(keys, indices)
        return top_k_vocab




"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.api._v2.keras.applications import convnext
from tf_keras.api._v2.keras.applications import densenet
from tf_keras.api._v2.keras.applications import efficientnet
from tf_keras.api._v2.keras.applications import efficientnet_v2
from tf_keras.api._v2.keras.applications import imagenet_utils
from tf_keras.api._v2.keras.applications import inception_resnet_v2
from tf_keras.api._v2.keras.applications import inception_v3
from tf_keras.api._v2.keras.applications import mobilenet
from tf_keras.api._v2.keras.applications import mobilenet_v2
from tf_keras.api._v2.keras.applications import mobilenet_v3
from tf_keras.api._v2.keras.applications import nasnet
from tf_keras.api._v2.keras.applications import regnet
from tf_keras.api._v2.keras.applications import resnet
from tf_keras.api._v2.keras.applications import resnet50
from tf_keras.api._v2.keras.applications import resnet_rs
from tf_keras.api._v2.keras.applications import resnet_v2
from tf_keras.api._v2.keras.applications import vgg16
from tf_keras.api._v2.keras.applications import vgg19
from tf_keras.api._v2.keras.applications import xception
from tf_keras.src.applications.convnext import ConvNeXtBase
from tf_keras.src.applications.convnext import ConvNeXtLarge
from tf_keras.src.applications.convnext import ConvNeXtSmall
from tf_keras.src.applications.convnext import ConvNeXtTiny
from tf_keras.src.applications.convnext import ConvNeXtXLarge
from tf_keras.src.applications.densenet import DenseNet121
from tf_keras.src.applications.densenet import DenseNet169
from tf_keras.src.applications.densenet import DenseNet201
from tf_keras.src.applications.efficientnet import EfficientNetB0
from tf_keras.src.applications.efficientnet import EfficientNetB1
from tf_keras.src.applications.efficientnet import EfficientNetB2
from tf_keras.src.applications.efficientnet import EfficientNetB3
from tf_keras.src.applications.efficientnet import EfficientNetB4
from tf_keras.src.applications.efficientnet import EfficientNetB5
from tf_keras.src.applications.efficientnet import EfficientNetB6
from tf_keras.src.applications.efficientnet import EfficientNetB7
from tf_keras.src.applications.efficientnet_v2 import EfficientNetV2B0
from tf_keras.src.applications.efficientnet_v2 import EfficientNetV2B1
from tf_keras.src.applications.efficientnet_v2 import EfficientNetV2B2
from tf_keras.src.applications.efficientnet_v2 import EfficientNetV2B3
from tf_keras.src.applications.efficientnet_v2 import EfficientNetV2L
from tf_keras.src.applications.efficientnet_v2 import EfficientNetV2M
from tf_keras.src.applications.efficientnet_v2 import EfficientNetV2S
from tf_keras.src.applications.inception_resnet_v2 import InceptionResNetV2
from tf_keras.src.applications.inception_v3 import InceptionV3
from tf_keras.src.applications.mobilenet import MobileNet
from tf_keras.src.applications.mobilenet_v2 import MobileNetV2
from tf_keras.src.applications.mobilenet_v3 import MobileNetV3Large
from tf_keras.src.applications.mobilenet_v3 import MobileNetV3Small
from tf_keras.src.applications.nasnet import NASNetLarge
from tf_keras.src.applications.nasnet import NASNetMobile
from tf_keras.src.applications.regnet import RegNetX002
from tf_keras.src.applications.regnet import RegNetX004
from tf_keras.src.applications.regnet import RegNetX006
from tf_keras.src.applications.regnet import RegNetX008
from tf_keras.src.applications.regnet import RegNetX016
from tf_keras.src.applications.regnet import RegNetX032
from tf_keras.src.applications.regnet import RegNetX040
from tf_keras.src.applications.regnet import RegNetX064
from tf_keras.src.applications.regnet import RegNetX080
from tf_keras.src.applications.regnet import RegNetX120
from tf_keras.src.applications.regnet import RegNetX160
from tf_keras.src.applications.regnet import RegNetX320
from tf_keras.src.applications.regnet import RegNetY002
from tf_keras.src.applications.regnet import RegNetY004
from tf_keras.src.applications.regnet import RegNetY006
from tf_keras.src.applications.regnet import RegNetY008
from tf_keras.src.applications.regnet import RegNetY016
from tf_keras.src.applications.regnet import RegNetY032
from tf_keras.src.applications.regnet import RegNetY040
from tf_keras.src.applications.regnet import RegNetY064
from tf_keras.src.applications.regnet import RegNetY080
from tf_keras.src.applications.regnet import RegNetY120
from tf_keras.src.applications.regnet import RegNetY160
from tf_keras.src.applications.regnet import RegNetY320
from tf_keras.src.applications.resnet import ResNet101
from tf_keras.src.applications.resnet import ResNet152
from tf_keras.src.applications.resnet import ResNet50
from tf_keras.src.applications.resnet_rs import ResNetRS101
from tf_keras.src.applications.resnet_rs import ResNetRS152
from tf_keras.src.applications.resnet_rs import ResNetRS200
from tf_keras.src.applications.resnet_rs import ResNetRS270
from tf_keras.src.applications.resnet_rs import ResNetRS350
from tf_keras.src.applications.resnet_rs import ResNetRS420
from tf_keras.src.applications.resnet_rs import ResNetRS50
from tf_keras.src.applications.resnet_v2 import ResNet101V2
from tf_keras.src.applications.resnet_v2 import ResNet152V2
from tf_keras.src.applications.resnet_v2 import ResNet50V2
from tf_keras.src.applications.vgg16 import VGG16
from tf_keras.src.applications.vgg19 import VGG19
from tf_keras.src.applications.xception import Xception



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Adam optimizer implementation."""

import tensorflow.compat.v2 as tf

from tf_keras.src.optimizers import optimizer
from tf_keras.src.saving.object_registration import register_keras_serializable

# isort: off
from tensorflow.python.util.tf_export import keras_export


@register_keras_serializable()
@keras_export(
    "keras.optimizers.Adam",
    "keras.optimizers.experimental.Adam",
    "keras.dtensor.experimental.optimizers.Adam",
    v1=[],
)
class Adam(optimizer.Optimizer):
    r"""Optimizer that implements the Adam algorithm.

    Adam optimization is a stochastic gradient descent method that is based on
    adaptive estimation of first-order and second-order moments.

    According to
    [Kingma et al., 2014](http://arxiv.org/abs/1412.6980),
    the method is "*computationally
    efficient, has little memory requirement, invariant to diagonal rescaling of
    gradients, and is well suited for problems that are large in terms of
    data/parameters*".

    Args:
        learning_rate: A `tf.Tensor`, floating point value, a schedule that is a
            `tf.keras.optimizers.schedules.LearningRateSchedule`, or a callable
            that takes no arguments and returns the actual value to use. The
            learning rate. Defaults to `0.001`.
        beta_1: A float value or a constant float tensor, or a callable
            that takes no arguments and returns the actual value to use. The
            exponential decay rate for the 1st moment estimates.
            Defaults to `0.9`.
        beta_2: A float value or a constant float tensor, or a callable
            that takes no arguments and returns the actual value to use. The
            exponential decay rate for the 2nd moment estimates.
            Defaults to `0.999`.
        epsilon: A small constant for numerical stability. This epsilon is
            "epsilon hat" in the Kingma and Ba paper (in the formula just before
            Section 2.1) if `adaptive_epsilon` is `False`, not the epsilon in
            Algorithm 1 of the paper. Defaults to `1e-7`.
        adaptive_epsilon: If `True` the epsilon hat is computed adaptively
            from the given epsilon (Algorithm 1 on the paper). If `False`, the
            epsilon given will be "epsilon hat". Default to `False`.
        amsgrad: Boolean. Whether to apply AMSGrad variant of this algorithm
            from the paper "On the Convergence of Adam and beyond".
            Defaults to `False`.
        {{base_optimizer_keyword_args}}

    Reference:
        - [Kingma et al., 2014](http://arxiv.org/abs/1412.6980)
        - [Reddi et al., 2018](
            https://openreview.net/pdf?id=ryQu7f-RZ) for `amsgrad`.

    Notes:

    The default value of 1e-7 for epsilon might not be a good default in
    general. For example, when training an Inception network on ImageNet a
    current good choice is 1.0 or 0.1. Note that since Adam uses the
    formulation just before Section 2.1 of the Kingma and Ba paper rather than
    the formulation in Algorithm 1, the "epsilon" referred to here is "epsilon
    hat" in the paper.

    The sparse implementation of this algorithm (used when the gradient is an
    IndexedSlices object, typically because of `tf.gather` or an embedding
    lookup in the forward pass) does apply momentum to variable slices even if
    they were not used in the forward pass (meaning they have a gradient equal
    to zero). Momentum decay (beta1) is also applied to the entire momentum
    accumulator. This means that the sparse behavior is equivalent to the dense
    behavior (in contrast to some momentum implementations which ignore momentum
    unless a variable slice was actually used).
    """

    def __init__(
        self,
        learning_rate=0.001,
        beta_1=0.9,
        beta_2=0.999,
        epsilon=1e-7,
        adaptive_epsilon=False,
        amsgrad=False,
        weight_decay=None,
        clipnorm=None,
        clipvalue=None,
        global_clipnorm=None,
        use_ema=False,
        ema_momentum=0.99,
        ema_overwrite_frequency=None,
        jit_compile=True,
        name="Adam",
        **kwargs
    ):
        super().__init__(
            name=name,
            weight_decay=weight_decay,
            clipnorm=clipnorm,
            clipvalue=clipvalue,
            global_clipnorm=global_clipnorm,
            use_ema=use_ema,
            ema_momentum=ema_momentum,
            ema_overwrite_frequency=ema_overwrite_frequency,
            jit_compile=jit_compile,
            **kwargs
        )
        self._learning_rate = self._build_learning_rate(learning_rate)
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon
        self.adaptive_epsilon = adaptive_epsilon
        self.amsgrad = amsgrad

    def build(self, var_list):
        """Initialize optimizer variables.

        Adam optimizer has 3 types of variables: momentums, velocities and
        velocity_hat (only set when amsgrad is applied),

        Args:
            var_list: list of model variables to build Adam variables on.
        """
        super().build(var_list)
        if hasattr(self, "_built") and self._built:
            return
        self._built = True
        self._momentums = []
        self._velocities = []
        for var in var_list:
            self._momentums.append(
                self.add_variable_from_reference(
                    model_variable=var, variable_name="m"
                )
            )
            self._velocities.append(
                self.add_variable_from_reference(
                    model_variable=var, variable_name="v"
                )
            )
        if self.amsgrad:
            self._velocity_hats = []
            for var in var_list:
                self._velocity_hats.append(
                    self.add_variable_from_reference(
                        model_variable=var, variable_name="vhat"
                    )
                )

    def update_step(self, gradient, variable):
        """Update step given gradient and the associated model variable."""
        lr = tf.cast(self.learning_rate, variable.dtype)
        local_step = tf.cast(self.iterations + 1, variable.dtype)
        beta_1_power = tf.pow(tf.cast(self.beta_1, variable.dtype), local_step)
        beta_2_power = tf.pow(tf.cast(self.beta_2, variable.dtype), local_step)

        var_key = self._var_key(variable)
        m = self._momentums[self._index_dict[var_key]]
        v = self._velocities[self._index_dict[var_key]]

        alpha = lr * tf.sqrt(1 - beta_2_power) / (1 - beta_1_power)
        if self.adaptive_epsilon:
            epsilon_hat = self.epsilon * tf.sqrt(1 - beta_2_power)
        else:
            epsilon_hat = self.epsilon

        if isinstance(gradient, tf.IndexedSlices):
            # Sparse gradients.
            m.assign_add(-m * (1 - self.beta_1))
            m.scatter_add(
                tf.IndexedSlices(
                    gradient.values * (1 - self.beta_1), gradient.indices
                )
            )
            v.assign_add(-v * (1 - self.beta_2))
            v.scatter_add(
                tf.IndexedSlices(
                    tf.square(gradient.values) * (1 - self.beta_2),
                    gradient.indices,
                )
            )
            if self.amsgrad:
                v_hat = self._velocity_hats[self._index_dict[var_key]]
                v_hat.assign(tf.maximum(v_hat, v))
                v = v_hat
            variable.assign_sub((m * alpha) / (tf.sqrt(v) + epsilon_hat))
        else:
            # Dense gradients.
            m.assign_add((gradient - m) * (1 - self.beta_1))
            v.assign_add((tf.square(gradient) - v) * (1 - self.beta_2))
            if self.amsgrad:
                v_hat = self._velocity_hats[self._index_dict[var_key]]
                v_hat.assign(tf.maximum(v_hat, v))
                v = v_hat
            variable.assign_sub((m * alpha) / (tf.sqrt(v) + epsilon_hat))

    def get_config(self):
        config = super().get_config()

        config.update(
            {
                "learning_rate": self._serialize_hyperparameter(
                    self._learning_rate
                ),
                "beta_1": self.beta_1,
                "beta_2": self.beta_2,
                "epsilon": self.epsilon,
                "amsgrad": self.amsgrad,
            }
        )
        return config


Adam.__doc__ = Adam.__doc__.replace(
    "{{base_optimizer_keyword_args}}", optimizer.base_optimizer_keyword_args
)




# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Benchmark for TF-Keras text vectorization preprocessing layer's adapt method.
"""

import os
import random
import string
import time

import numpy as np
import tensorflow.compat.v2 as tf

import tf_keras.src as keras
from tf_keras.src.layers.preprocessing import index_lookup


# word_gen creates random sequences of ASCII letters (both lowercase and upper).
# The number of unique strings is ~2,700.
def tensor_gen(batch, num_elements):
    data = []
    for _ in range(batch):
        batch_element = []
        for _ in range(num_elements - 1):
            tok = "".join(random.choice(string.ascii_letters) for i in range(2))
            batch_element.append(tok)
        batch_element.append("")  # Explicitly test the empty string.
        data.append(batch_element)
    return tf.constant(data)


def get_vocab():
    vocab = list(
        set([a + b for a in string.ascii_letters for b in string.ascii_letters])
    )
    vocab.sort()
    return vocab


# This class uses TestCase for get_temp_dir().
class BenchmarkLookup(tf.test.Benchmark):
    """Benchmark the index lookup layer's forward pass."""

    def _write_to_temp_file(self, file_name, vocab_list):
        vocab_path = os.path.join(self.get_temp_dir(), file_name + ".txt")
        with tf.io.gfile.GFile(vocab_path, "w") as writer:
            for vocab in vocab_list:
                writer.write(vocab + "\n")
            writer.flush()
            writer.close()
        return vocab_path

    def run_numpy_implementation(self, data, vocab):
        """Test the python implementation."""
        input_t = keras.Input(shape=(), dtype=tf.string)
        layer = index_lookup.IndexLookup(
            vocabulary=vocab,
            max_tokens=None,
            num_oov_indices=1,
            mask_token="",
            oov_token="OOV",
            dtype=tf.string,
        )
        out_t = layer(input_t)
        model = keras.Model(input_t, out_t)
        num_repeats = 5
        starts = []
        ends = []
        _ = model(data)
        for _ in range(num_repeats):
            starts.append(time.time())
            out = model(data)
            ends.append(time.time())
        avg_time = np.mean(np.array(ends) - np.array(starts))
        return avg_time, out

    def bm_adapt_implementation(self, num_elements, batch_size):
        """Test the KPL adapt implementation."""
        vocab = get_vocab()
        vocab_file = self._write_to_temp_file("vocab", vocab)
        vocabulary_initializer = tf.lookup.TextFileInitializer(
            filename=vocab_file,
            key_dtype=tf.string,
            key_index=tf.lookup.TextFileIndex.WHOLE_LINE,
            value_dtype=tf.int64,
            value_index=tf.lookup.TextFileIndex.LINE_NUMBER,
            value_index_offset=2,
        )
        input_t = keras.Input(shape=(), dtype=tf.string)
        layer = index_lookup.IndexLookup(
            vocabulary=vocabulary_initializer,
            max_tokens=None,
            num_oov_indices=1,
            mask_token="",
            oov_token="OOV",
            dtype=tf.string,
        )
        out_t = layer(input_t)
        model = keras.Model(input_t, out_t)
        num_repeats = 5
        starts = []
        ends = []
        data = tensor_gen(batch_size, num_elements)
        _ = model(data)
        for _ in range(num_repeats):
            starts.append(time.time())
            _ = model(data)
            ends.append(time.time())
        avg_time = np.mean(np.array(ends) - np.array(starts))
        baseline, _ = self.run_numpy_implementation(data, vocab)
        extras = {
            "numpy implementation baseline": baseline,
            "delta seconds": (baseline - avg_time),
            "delta percent": ((baseline - avg_time) / baseline) * 100,
        }
        name = "index_lookup_forward|%s_elements|batch_%s" % (
            num_elements,
            batch_size,
        )
        self.report_benchmark(
            iters=num_repeats, wall_time=avg_time, extras=extras, name=name
        )

    def benchmark_vocab_size_by_batch(self):
        for tensor_size in [100, 1000, 10000]:
            for batch in [1, 16, 2048]:
                self.bm_adapt_implementation(tensor_size, batch)


if __name__ == "__main__":
    tf.test.main()




# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""CIFAR100 small images classification dataset."""

import os

import numpy as np

from tf_keras.src import backend
from tf_keras.src.datasets.cifar import load_batch
from tf_keras.src.utils.data_utils import get_file

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.datasets.cifar100.load_data")
def load_data(label_mode="fine", cache_dir=None):
    """Loads the CIFAR100 dataset.

    This is a dataset of 50,000 32x32 color training images and
    10,000 test images, labeled over 100 fine-grained classes that are
    grouped into 20 coarse-grained classes. See more info at the
    [CIFAR homepage](https://www.cs.toronto.edu/~kriz/cifar.html).

    Args:
      label_mode: one of "fine", "coarse". If it is "fine" the category labels
        are the fine-grained labels, if it is "coarse" the output labels are the
        coarse-grained superclasses.
      cache_dir: directory where to cache the dataset locally. When None,
        defaults to `~/.keras/datasets`.

    Returns:
      Tuple of NumPy arrays: `(x_train, y_train), (x_test, y_test)`.

    **x_train**: uint8 NumPy array of image data with shapes
      `(50000, 32, 32, 3)`, containing the training data. Pixel values range
      from 0 to 255.

    **y_train**: uint8 NumPy array of labels (integers in range 0-99)
      with shape `(50000, 1)` for the training data.

    **x_test**: uint8 NumPy array of image data with shapes
      `(10000, 32, 32, 3)`, containing the test data. Pixel values range
      from 0 to 255.

    **y_test**: uint8 NumPy array of labels (integers in range 0-99)
      with shape `(10000, 1)` for the test data.

    Example:

    ```python
    (x_train, y_train), (x_test, y_test) = keras.datasets.cifar100.load_data()
    assert x_train.shape == (50000, 32, 32, 3)
    assert x_test.shape == (10000, 32, 32, 3)
    assert y_train.shape == (50000, 1)
    assert y_test.shape == (10000, 1)
    ```
    """
    if label_mode not in ["fine", "coarse"]:
        raise ValueError(
            '`label_mode` must be one of `"fine"`, `"coarse"`. '
            f"Received: label_mode={label_mode}."
        )

    dirname = "cifar-100-python"
    origin = "https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz"
    if cache_dir:
        cache_dir = os.path.expanduser(cache_dir)
        os.makedirs(cache_dir, exist_ok=True)
    path = get_file(
        dirname,
        origin=origin,
        untar=True,
        file_hash=(  # noqa: E501
            "85cd44d02ba6437773c5bbd22e183051d648de2e7d6b014e1ef29b855ba677a7"
        ),
        cache_dir=cache_dir,
    )

    fpath = os.path.join(path, "train")
    x_train, y_train = load_batch(fpath, label_key=label_mode + "_labels")

    fpath = os.path.join(path, "test")
    x_test, y_test = load_batch(fpath, label_key=label_mode + "_labels")

    y_train = np.reshape(y_train, (len(y_train), 1))
    y_test = np.reshape(y_test, (len(y_test), 1))

    if backend.image_data_format() == "channels_last":
        x_train = x_train.transpose(0, 2, 3, 1)
        x_test = x_test.transpose(0, 2, 3, 1)

    return (x_train, y_train), (x_test, y_test)




"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.applications.nasnet import NASNetLarge
from tf_keras.src.applications.nasnet import NASNetMobile
from tf_keras.src.applications.nasnet import decode_predictions
from tf_keras.src.applications.nasnet import preprocess_input



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.engine.base_preprocessing_layer import PreprocessingLayer
from tf_keras.src.layers.preprocessing.category_encoding import CategoryEncoding
from tf_keras.src.layers.preprocessing.discretization import Discretization
from tf_keras.src.layers.preprocessing.hashed_crossing import HashedCrossing
from tf_keras.src.layers.preprocessing.hashing import Hashing
from tf_keras.src.layers.preprocessing.image_preprocessing import CenterCrop
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomContrast
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomCrop
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomFlip
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomHeight
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomRotation
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomTranslation
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomWidth
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomZoom
from tf_keras.src.layers.preprocessing.image_preprocessing import Rescaling
from tf_keras.src.layers.preprocessing.image_preprocessing import Resizing
from tf_keras.src.layers.preprocessing.integer_lookup import IntegerLookup
from tf_keras.src.layers.preprocessing.normalization import Normalization
from tf_keras.src.layers.preprocessing.string_lookup import StringLookup
from tf_keras.src.layers.preprocessing.text_vectorization import TextVectorization



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.applications.densenet import DenseNet121
from tf_keras.src.applications.densenet import DenseNet169
from tf_keras.src.applications.densenet import DenseNet201
from tf_keras.src.applications.densenet import decode_predictions
from tf_keras.src.applications.densenet import preprocess_input



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Built-in optimizer classes.

For more examples see the base class `tf.keras.optimizers.Optimizer`.
"""

# Imports needed for deserialization.

import platform
import warnings

import tensorflow.compat.v2 as tf
from absl import logging

from tf_keras.src import backend
from tf_keras.src.optimizers import adadelta
from tf_keras.src.optimizers import adafactor
from tf_keras.src.optimizers import adagrad
from tf_keras.src.optimizers import adam
from tf_keras.src.optimizers import adamax
from tf_keras.src.optimizers import adamw
from tf_keras.src.optimizers import ftrl
from tf_keras.src.optimizers import lion
from tf_keras.src.optimizers import nadam
from tf_keras.src.optimizers import optimizer as base_optimizer
from tf_keras.src.optimizers import rmsprop
from tf_keras.src.optimizers import sgd
from tf_keras.src.optimizers.legacy import adadelta as adadelta_legacy
from tf_keras.src.optimizers.legacy import adagrad as adagrad_legacy
from tf_keras.src.optimizers.legacy import adam as adam_legacy
from tf_keras.src.optimizers.legacy import adamax as adamax_legacy
from tf_keras.src.optimizers.legacy import ftrl as ftrl_legacy
from tf_keras.src.optimizers.legacy import (
    gradient_descent as gradient_descent_legacy,
)
from tf_keras.src.optimizers.legacy import nadam as nadam_legacy
from tf_keras.src.optimizers.legacy import optimizer_v2 as base_optimizer_legacy
from tf_keras.src.optimizers.legacy import rmsprop as rmsprop_legacy
from tf_keras.src.optimizers.legacy.adadelta import Adadelta
from tf_keras.src.optimizers.legacy.adagrad import Adagrad
from tf_keras.src.optimizers.legacy.adam import Adam
from tf_keras.src.optimizers.legacy.adamax import Adamax
from tf_keras.src.optimizers.legacy.ftrl import Ftrl

# Symbols to be accessed under keras.optimizers. To be replaced with
# optimizers v2022 when they graduate out of experimental.
from tf_keras.src.optimizers.legacy.gradient_descent import SGD
from tf_keras.src.optimizers.legacy.nadam import Nadam
from tf_keras.src.optimizers.legacy.rmsprop import RMSprop
from tf_keras.src.optimizers.optimizer_v1 import Optimizer
from tf_keras.src.optimizers.optimizer_v1 import TFOptimizer
from tf_keras.src.optimizers.schedules import learning_rate_schedule
from tf_keras.src.saving.legacy import serialization as legacy_serialization
from tf_keras.src.saving.serialization_lib import deserialize_keras_object
from tf_keras.src.saving.serialization_lib import serialize_keras_object

# isort: off
from tensorflow.python.util.tf_export import keras_export

# pylint: disable=line-too-long


@keras_export("keras.optimizers.serialize")
def serialize(optimizer, use_legacy_format=False):
    """Serialize the optimizer configuration to JSON compatible python dict.

    The configuration can be used for persistence and reconstruct the
    `Optimizer` instance again.

    >>> tf.keras.optimizers.serialize(tf.keras.optimizers.legacy.SGD())
    {'module': 'keras.optimizers.legacy', 'class_name': 'SGD', 'config': {'name': 'SGD', 'learning_rate': 0.01, 'decay': 0.0, 'momentum': 0.0, 'nesterov': False}, 'registered_name': None}
    """  # noqa: E501
    """
    Args:
      optimizer: An `Optimizer` instance to serialize.

    Returns:
      Python dict which contains the configuration of the input optimizer.
    """
    if optimizer is None:
        return None
    if not isinstance(
        optimizer,
        (
            base_optimizer.Optimizer,
            Optimizer,
            base_optimizer_legacy.OptimizerV2,
        ),
    ):
        warnings.warn(
            "The `keras.optimizers.serialize()` API should only be used for "
            "objects of type `keras.optimizers.Optimizer`. Found an instance "
            f"of type {type(optimizer)}, which may lead to improper "
            "serialization."
        )
    if use_legacy_format:
        return legacy_serialization.serialize_keras_object(optimizer)
    return serialize_keras_object(optimizer)


def is_arm_mac():
    return platform.system() == "Darwin" and platform.processor() == "arm"


@keras_export("keras.optimizers.deserialize")
def deserialize(config, custom_objects=None, use_legacy_format=False, **kwargs):
    """Inverse of the `serialize` function.

    Args:
        config: Optimizer configuration dictionary.
        custom_objects: Optional dictionary mapping names (strings) to custom
            objects (classes and functions) to be considered during
            deserialization.

    Returns:
        A TF-Keras Optimizer instance.
    """
    # loss_scale_optimizer has a direct dependency of optimizer, import here
    # rather than top to avoid the cyclic dependency.
    from tf_keras.src.mixed_precision import (
        loss_scale_optimizer,
    )

    use_legacy_optimizer = kwargs.pop("use_legacy_optimizer", False)
    if kwargs:
        raise TypeError(f"Invalid keyword arguments: {kwargs}")
    if len(config["config"]) > 0:
        # If the optimizer config is not empty, then we use the value of
        # `is_legacy_optimizer` to override `use_legacy_optimizer`. If
        # `is_legacy_optimizer` does not exist in config, it means we are
        # using the legacy optimzier.
        use_legacy_optimizer = config["config"].get("is_legacy_optimizer", True)
    if (
        tf.__internal__.tf2.enabled()
        and tf.executing_eagerly()
        and not is_arm_mac()
        and not use_legacy_optimizer
    ):
        # We observed a slowdown of optimizer on M1 Mac, so we fall back to the
        # legacy optimizer for M1 users now, see b/263339144 for more context.
        all_classes = {
            "adadelta": adadelta.Adadelta,
            "adagrad": adagrad.Adagrad,
            "adam": adam.Adam,
            "adamw": adamw.AdamW,
            "adamax": adamax.Adamax,
            "experimentaladadelta": adadelta.Adadelta,
            "experimentaladagrad": adagrad.Adagrad,
            "experimentaladam": adam.Adam,
            "experimentalsgd": sgd.SGD,
            "nadam": nadam.Nadam,
            "rmsprop": rmsprop.RMSprop,
            "sgd": sgd.SGD,
            "ftrl": ftrl.Ftrl,
            "lossscaleoptimizer": loss_scale_optimizer.LossScaleOptimizerV3,
            "lossscaleoptimizerv3": loss_scale_optimizer.LossScaleOptimizerV3,
            # LossScaleOptimizerV1 was an old version of LSO that was removed.
            # Deserializing it turns it into a LossScaleOptimizer
            "lossscaleoptimizerv1": loss_scale_optimizer.LossScaleOptimizer,
        }
    else:
        all_classes = {
            "adadelta": adadelta_legacy.Adadelta,
            "adagrad": adagrad_legacy.Adagrad,
            "adam": adam_legacy.Adam,
            "adamax": adamax_legacy.Adamax,
            "experimentaladadelta": adadelta.Adadelta,
            "experimentaladagrad": adagrad.Adagrad,
            "experimentaladam": adam.Adam,
            "experimentalsgd": sgd.SGD,
            "nadam": nadam_legacy.Nadam,
            "rmsprop": rmsprop_legacy.RMSprop,
            "sgd": gradient_descent_legacy.SGD,
            "ftrl": ftrl_legacy.Ftrl,
            "lossscaleoptimizer": loss_scale_optimizer.LossScaleOptimizer,
            "lossscaleoptimizerv3": loss_scale_optimizer.LossScaleOptimizerV3,
            # LossScaleOptimizerV1 was an old version of LSO that was removed.
            # Deserializing it turns it into a LossScaleOptimizer
            "lossscaleoptimizerv1": loss_scale_optimizer.LossScaleOptimizer,
        }

    # Make deserialization case-insensitive for built-in optimizers.
    if config["class_name"].lower() in all_classes:
        config["class_name"] = config["class_name"].lower()

    if use_legacy_format:
        return legacy_serialization.deserialize_keras_object(
            config,
            module_objects=all_classes,
            custom_objects=custom_objects,
            printable_module_name="optimizer",
        )

    return deserialize_keras_object(
        config,
        module_objects=all_classes,
        custom_objects=custom_objects,
        printable_module_name="optimizer",
    )


@keras_export(
    "keras.__internal__.optimizers.convert_to_legacy_optimizer", v1=[]
)
def convert_to_legacy_optimizer(optimizer):
    """Convert experimental optimizer to legacy optimizer.

    This function takes in a `keras.optimizers.Optimizer`
    instance and converts it to the corresponding
    `keras.optimizers.legacy.Optimizer` instance.
    For example, `keras.optimizers.Adam(...)` to
    `keras.optimizers.legacy.Adam(...)`.

    Args:
        optimizer: An instance of `keras.optimizers.Optimizer`.
    """
    # loss_scale_optimizer has a direct dependency of optimizer, import here
    # rather than top to avoid the cyclic dependency.
    from tf_keras.src.mixed_precision import (
        loss_scale_optimizer,
    )

    if not isinstance(optimizer, base_optimizer.Optimizer):
        raise ValueError(
            "`convert_to_legacy_optimizer` should only be called "
            "on instances of `tf.keras.optimizers.Optimizer`, but "
            f"received {optimizer} of type {type(optimizer)}."
        )
    optimizer_name = optimizer.__class__.__name__.lower()
    config = optimizer.get_config()
    # Remove fields that only exist in experimental optimizer.
    keys_to_remove = [
        "weight_decay",
        "use_ema",
        "ema_momentum",
        "ema_overwrite_frequency",
        "jit_compile",
        "is_legacy_optimizer",
    ]
    for key in keys_to_remove:
        config.pop(key, None)

    if isinstance(optimizer, loss_scale_optimizer.LossScaleOptimizerV3):
        # For LossScaleOptimizers, recursively convert the inner optimizer
        config["inner_optimizer"] = convert_to_legacy_optimizer(
            optimizer.inner_optimizer
        )
        if optimizer_name == "lossscaleoptimizerv3":
            optimizer_name = "lossscaleoptimizer"

    # Learning rate can be a custom LearningRateSchedule, which is stored as
    # a dict in config, and cannot be deserialized.
    if hasattr(optimizer, "_learning_rate") and isinstance(
        optimizer._learning_rate, learning_rate_schedule.LearningRateSchedule
    ):
        config["learning_rate"] = optimizer._learning_rate
    legacy_optimizer_config = {
        "class_name": optimizer_name,
        "config": config,
    }
    return deserialize(
        legacy_optimizer_config,
        use_legacy_optimizer=True,
        use_legacy_format=True,
    )


@keras_export("keras.optimizers.get")
def get(identifier, **kwargs):
    """Retrieves a TF-Keras Optimizer instance.

    Args:
        identifier: Optimizer identifier, one of - String: name of an optimizer
          - Dictionary: configuration dictionary. - TF-Keras Optimizer instance
          (it will be returned unchanged). - TensorFlow Optimizer instance (it
          will be wrapped as a TF-Keras Optimizer).

    Returns:
        A TF-Keras Optimizer instance.

    Raises:
        ValueError: If `identifier` cannot be interpreted.
    """
    use_legacy_optimizer = kwargs.pop("use_legacy_optimizer", False)
    if kwargs:
        raise TypeError(f"Invalid keyword arguments: {kwargs}")
    if isinstance(
        identifier,
        (
            Optimizer,
            base_optimizer_legacy.OptimizerV2,
        ),
    ):
        return identifier
    elif isinstance(identifier, base_optimizer.Optimizer):
        if tf.__internal__.tf2.enabled():
            return identifier
        else:
            # If TF2 is disabled, we convert to the legacy
            # optimizer.
            return convert_to_legacy_optimizer(identifier)

    # Wrap legacy TF optimizer instances
    elif isinstance(identifier, tf.compat.v1.train.Optimizer):
        opt = TFOptimizer(identifier)
        backend.track_tf_optimizer(opt)
        return opt
    elif isinstance(identifier, dict):
        use_legacy_format = "module" not in identifier
        return deserialize(
            identifier,
            use_legacy_optimizer=use_legacy_optimizer,
            use_legacy_format=use_legacy_format,
        )
    elif isinstance(identifier, str):
        config = {"class_name": str(identifier), "config": {}}
        return get(
            config,
            use_legacy_optimizer=use_legacy_optimizer,
        )
    else:
        raise ValueError(
            f"Could not interpret optimizer identifier: {identifier}"
        )




"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.engine.base_preprocessing_layer import PreprocessingLayer
from tf_keras.src.layers.preprocessing.category_encoding import CategoryEncoding
from tf_keras.src.layers.preprocessing.discretization import Discretization
from tf_keras.src.layers.preprocessing.hashing import Hashing
from tf_keras.src.layers.preprocessing.image_preprocessing import CenterCrop
from tf_keras.src.layers.preprocessing.image_preprocessing import Rescaling
from tf_keras.src.layers.preprocessing.image_preprocessing import Resizing
from tf_keras.src.layers.preprocessing.normalization import Normalization



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras merging layers."""


# Merging functions.
# Merging layers.
from tf_keras.src.layers.merging.add import Add
from tf_keras.src.layers.merging.add import add
from tf_keras.src.layers.merging.average import Average
from tf_keras.src.layers.merging.average import average
from tf_keras.src.layers.merging.concatenate import Concatenate
from tf_keras.src.layers.merging.concatenate import concatenate
from tf_keras.src.layers.merging.dot import Dot
from tf_keras.src.layers.merging.dot import dot
from tf_keras.src.layers.merging.maximum import Maximum
from tf_keras.src.layers.merging.maximum import maximum
from tf_keras.src.layers.merging.minimum import Minimum
from tf_keras.src.layers.merging.minimum import minimum
from tf_keras.src.layers.merging.multiply import Multiply
from tf_keras.src.layers.merging.multiply import multiply
from tf_keras.src.layers.merging.subtract import Subtract
from tf_keras.src.layers.merging.subtract import subtract




"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras import __internal__
from tf_keras import activations
from tf_keras import applications
from tf_keras import backend
from tf_keras import callbacks
from tf_keras import constraints
from tf_keras import datasets
from tf_keras import dtensor
from tf_keras import estimator
from tf_keras import experimental
from tf_keras import export
from tf_keras import initializers
from tf_keras import layers
from tf_keras import losses
from tf_keras import metrics
from tf_keras import mixed_precision
from tf_keras import models
from tf_keras import optimizers
from tf_keras import preprocessing
from tf_keras import regularizers
from tf_keras import saving
from tf_keras import utils
from tf_keras.src.engine.input_layer import Input
from tf_keras.src.engine.sequential import Sequential
from tf_keras.src.engine.training import Model


__version__ = "2.20.0.dev2025041209"



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.api._v1.keras.layers import experimental
from tf_keras.src.engine.base_layer import Layer
from tf_keras.src.engine.base_layer_utils import disable_v2_dtype_behavior
from tf_keras.src.engine.base_layer_utils import enable_v2_dtype_behavior
from tf_keras.src.engine.input_layer import Input
from tf_keras.src.engine.input_layer import InputLayer
from tf_keras.src.engine.input_spec import InputSpec
from tf_keras.src.feature_column.dense_features import DenseFeatures
from tf_keras.src.layers.activation.elu import ELU
from tf_keras.src.layers.activation.leaky_relu import LeakyReLU
from tf_keras.src.layers.activation.prelu import PReLU
from tf_keras.src.layers.activation.relu import ReLU
from tf_keras.src.layers.activation.softmax import Softmax
from tf_keras.src.layers.activation.thresholded_relu import ThresholdedReLU
from tf_keras.src.layers.attention.additive_attention import AdditiveAttention
from tf_keras.src.layers.attention.attention import Attention
from tf_keras.src.layers.attention.multi_head_attention import MultiHeadAttention
from tf_keras.src.layers.convolutional.conv1d import Conv1D
from tf_keras.src.layers.convolutional.conv1d import Conv1D as Convolution1D
from tf_keras.src.layers.convolutional.conv1d_transpose import Conv1DTranspose
from tf_keras.src.layers.convolutional.conv1d_transpose import Conv1DTranspose as Convolution1DTranspose
from tf_keras.src.layers.convolutional.conv2d import Conv2D
from tf_keras.src.layers.convolutional.conv2d import Conv2D as Convolution2D
from tf_keras.src.layers.convolutional.conv2d_transpose import Conv2DTranspose
from tf_keras.src.layers.convolutional.conv2d_transpose import Conv2DTranspose as Convolution2DTranspose
from tf_keras.src.layers.convolutional.conv3d import Conv3D
from tf_keras.src.layers.convolutional.conv3d import Conv3D as Convolution3D
from tf_keras.src.layers.convolutional.conv3d_transpose import Conv3DTranspose
from tf_keras.src.layers.convolutional.conv3d_transpose import Conv3DTranspose as Convolution3DTranspose
from tf_keras.src.layers.convolutional.depthwise_conv1d import DepthwiseConv1D
from tf_keras.src.layers.convolutional.depthwise_conv2d import DepthwiseConv2D
from tf_keras.src.layers.convolutional.separable_conv1d import SeparableConv1D
from tf_keras.src.layers.convolutional.separable_conv1d import SeparableConv1D as SeparableConvolution1D
from tf_keras.src.layers.convolutional.separable_conv2d import SeparableConv2D
from tf_keras.src.layers.convolutional.separable_conv2d import SeparableConv2D as SeparableConvolution2D
from tf_keras.src.layers.core.activation import Activation
from tf_keras.src.layers.core.dense import Dense
from tf_keras.src.layers.core.einsum_dense import EinsumDense
from tf_keras.src.layers.core.embedding import Embedding
from tf_keras.src.layers.core.identity import Identity
from tf_keras.src.layers.core.lambda_layer import Lambda
from tf_keras.src.layers.core.masking import Masking
from tf_keras.src.layers.locally_connected.locally_connected1d import LocallyConnected1D
from tf_keras.src.layers.locally_connected.locally_connected2d import LocallyConnected2D
from tf_keras.src.layers.merging.add import Add
from tf_keras.src.layers.merging.add import add
from tf_keras.src.layers.merging.average import Average
from tf_keras.src.layers.merging.average import average
from tf_keras.src.layers.merging.concatenate import Concatenate
from tf_keras.src.layers.merging.concatenate import concatenate
from tf_keras.src.layers.merging.dot import Dot
from tf_keras.src.layers.merging.dot import dot
from tf_keras.src.layers.merging.maximum import Maximum
from tf_keras.src.layers.merging.maximum import maximum
from tf_keras.src.layers.merging.minimum import Minimum
from tf_keras.src.layers.merging.minimum import minimum
from tf_keras.src.layers.merging.multiply import Multiply
from tf_keras.src.layers.merging.multiply import multiply
from tf_keras.src.layers.merging.subtract import Subtract
from tf_keras.src.layers.merging.subtract import subtract
from tf_keras.src.layers.normalization.batch_normalization_v1 import BatchNormalization
from tf_keras.src.layers.normalization.layer_normalization import LayerNormalization
from tf_keras.src.layers.pooling.average_pooling1d import AveragePooling1D
from tf_keras.src.layers.pooling.average_pooling1d import AveragePooling1D as AvgPool1D
from tf_keras.src.layers.pooling.average_pooling2d import AveragePooling2D
from tf_keras.src.layers.pooling.average_pooling2d import AveragePooling2D as AvgPool2D
from tf_keras.src.layers.pooling.average_pooling3d import AveragePooling3D
from tf_keras.src.layers.pooling.average_pooling3d import AveragePooling3D as AvgPool3D
from tf_keras.src.layers.pooling.global_average_pooling1d import GlobalAveragePooling1D
from tf_keras.src.layers.pooling.global_average_pooling1d import GlobalAveragePooling1D as GlobalAvgPool1D
from tf_keras.src.layers.pooling.global_average_pooling2d import GlobalAveragePooling2D
from tf_keras.src.layers.pooling.global_average_pooling2d import GlobalAveragePooling2D as GlobalAvgPool2D
from tf_keras.src.layers.pooling.global_average_pooling3d import GlobalAveragePooling3D
from tf_keras.src.layers.pooling.global_average_pooling3d import GlobalAveragePooling3D as GlobalAvgPool3D
from tf_keras.src.layers.pooling.global_max_pooling1d import GlobalMaxPooling1D
from tf_keras.src.layers.pooling.global_max_pooling1d import GlobalMaxPooling1D as GlobalMaxPool1D
from tf_keras.src.layers.pooling.global_max_pooling2d import GlobalMaxPooling2D
from tf_keras.src.layers.pooling.global_max_pooling2d import GlobalMaxPooling2D as GlobalMaxPool2D
from tf_keras.src.layers.pooling.global_max_pooling3d import GlobalMaxPooling3D
from tf_keras.src.layers.pooling.global_max_pooling3d import GlobalMaxPooling3D as GlobalMaxPool3D
from tf_keras.src.layers.pooling.max_pooling1d import MaxPooling1D
from tf_keras.src.layers.pooling.max_pooling1d import MaxPooling1D as MaxPool1D
from tf_keras.src.layers.pooling.max_pooling2d import MaxPooling2D
from tf_keras.src.layers.pooling.max_pooling2d import MaxPooling2D as MaxPool2D
from tf_keras.src.layers.pooling.max_pooling3d import MaxPooling3D
from tf_keras.src.layers.pooling.max_pooling3d import MaxPooling3D as MaxPool3D
from tf_keras.src.layers.preprocessing.category_encoding import CategoryEncoding
from tf_keras.src.layers.preprocessing.discretization import Discretization
from tf_keras.src.layers.preprocessing.hashing import Hashing
from tf_keras.src.layers.preprocessing.image_preprocessing import CenterCrop
from tf_keras.src.layers.preprocessing.image_preprocessing import Rescaling
from tf_keras.src.layers.preprocessing.image_preprocessing import Resizing
from tf_keras.src.layers.preprocessing.normalization import Normalization
from tf_keras.src.layers.regularization.activity_regularization import ActivityRegularization
from tf_keras.src.layers.regularization.alpha_dropout import AlphaDropout
from tf_keras.src.layers.regularization.dropout import Dropout
from tf_keras.src.layers.regularization.gaussian_dropout import GaussianDropout
from tf_keras.src.layers.regularization.gaussian_noise import GaussianNoise
from tf_keras.src.layers.regularization.spatial_dropout1d import SpatialDropout1D
from tf_keras.src.layers.regularization.spatial_dropout2d import SpatialDropout2D
from tf_keras.src.layers.regularization.spatial_dropout3d import SpatialDropout3D
from tf_keras.src.layers.reshaping.cropping1d import Cropping1D
from tf_keras.src.layers.reshaping.cropping2d import Cropping2D
from tf_keras.src.layers.reshaping.cropping3d import Cropping3D
from tf_keras.src.layers.reshaping.flatten import Flatten
from tf_keras.src.layers.reshaping.permute import Permute
from tf_keras.src.layers.reshaping.repeat_vector import RepeatVector
from tf_keras.src.layers.reshaping.reshape import Reshape
from tf_keras.src.layers.reshaping.up_sampling1d import UpSampling1D
from tf_keras.src.layers.reshaping.up_sampling2d import UpSampling2D
from tf_keras.src.layers.reshaping.up_sampling3d import UpSampling3D
from tf_keras.src.layers.reshaping.zero_padding1d import ZeroPadding1D
from tf_keras.src.layers.reshaping.zero_padding2d import ZeroPadding2D
from tf_keras.src.layers.reshaping.zero_padding3d import ZeroPadding3D
from tf_keras.src.layers.rnn.abstract_rnn_cell import AbstractRNNCell
from tf_keras.src.layers.rnn.base_rnn import RNN
from tf_keras.src.layers.rnn.base_wrapper import Wrapper
from tf_keras.src.layers.rnn.bidirectional import Bidirectional
from tf_keras.src.layers.rnn.conv_lstm1d import ConvLSTM1D
from tf_keras.src.layers.rnn.conv_lstm2d import ConvLSTM2D
from tf_keras.src.layers.rnn.conv_lstm3d import ConvLSTM3D
from tf_keras.src.layers.rnn.cudnn_gru import CuDNNGRU
from tf_keras.src.layers.rnn.cudnn_lstm import CuDNNLSTM
from tf_keras.src.layers.rnn.gru_v1 import GRU
from tf_keras.src.layers.rnn.gru_v1 import GRUCell
from tf_keras.src.layers.rnn.lstm_v1 import LSTM
from tf_keras.src.layers.rnn.lstm_v1 import LSTMCell
from tf_keras.src.layers.rnn.simple_rnn import SimpleRNN
from tf_keras.src.layers.rnn.simple_rnn import SimpleRNNCell
from tf_keras.src.layers.rnn.stacked_rnn_cells import StackedRNNCells
from tf_keras.src.layers.rnn.time_distributed import TimeDistributed
from tf_keras.src.layers.serialization import deserialize
from tf_keras.src.layers.serialization import serialize
